"""
Provides helper functions when working with text.

This module is more concerned with bytes/unicode and text processing, compared with string.py which
is used for string manipulation.

Functions in this file are copied or modified from Chapter 4 of:
        Fluent Python, 2nd ed., by Luciano Ramalho (O'Reilly).
        Copyright 2022 Luciano Ramalho, 978-1-492-05635-5
"""
from unicodedata import normalize, combining
import string


def normalize_equal(str1: str, str2: str, case_fold: bool = True, method: str = 'NFC') -> bool:
    """
    Compares two strings for equality after normalizing by the given `method`.

    "NFC is the normalization form recommended by the W3C [... and] make comparisons work as
    expected." (Fluent Python, pg 140.)

    "NFKC [normalziation ...] may lose or distort information but [it] can produce convenient
    intermediate representations for searching and indexes". (Fluent Python, pg 140.)

    This function is modified from:
        Fluent Python, 2nd ed., by Luciano Ramalho (O'Reilly). Pg. 144
        Copyright 2022 Luciano Ramalho, 978-1-492-05635-5

    >>> (normalize('NFC', 'café'), normalize('NFC', 'cafe\u0301'))
    ('café', 'café')
    >>> (len('café'), len('cafe\u0301'))
    (4, 5)
    >>> 'café' == 'cafe\u0301'
    False
    >>> normalize_equal('café', 'cafe\u0301')
    True
    >>> normalize_equal('café', 'Cafe\u0301')
    True
    >>> normalize_equal('café', 'Cafe\u0301', case_fold=False)
    False

    >>> ohm = '\u2126'
    >>> omega = 'Ω'
    >>> (ohm, omega)
    ('Ω', 'Ω')
    >>> (normalize('NFC', ohm), normalize('NFC', omega))
    ('Ω', 'Ω')
    >>> ohm == omega
    False
    >>> normalize_equal(ohm, omega)
    True
    >>> normalize_equal(ohm, omega, case_fold=False)
    True

    >>> normalize('NFC', '½')
    '½'
    >>> normalize('NFKC', '½')
    '1⁄2'
    >>> '1⁄2' == '1/2'
    False

    >>> normalize('NFC', 'Straße').casefold()
    'strasse'
    >>> normalize_equal('Straße', 'strasse', case_fold=False)
    False
    >>> normalize_equal('Straße', 'strasse')
    True

    Args:
        str1: the first string value
        str2: the second string value
        case_fold: if True, apply case folding (essentially converting text to lower case)
        method: the method; valid values are 'NFC', 'NFD', 'NFKC', 'NFKD'
    """  # noqa: RUF002
    if case_fold:
        return (normalize(method, str1).casefold() == normalize(method, str2).casefold())

    return normalize(method, str1) == normalize(method, str2)


def remove_marks(text: str, latin_only: bool = True) -> str:
    """
    Remove all diacritic marks.

    This function and the examples below are copied/modified from:
        Fluent Python, 2nd ed., by Luciano Ramalho (O'Reilly). Pg. 145-146
        Copyright 2022 Luciano Ramalho, 978-1-492-05635-5

    >>> order = '“Herr Voß: • ½ cup of Œtker™ caffè latte • bowl of açaí.”'
    >>> remove_marks(order, latin_only=False)
    '“Herr Voß: • ½ cup of Œtker™ caffe latte • bowl of acai.”'
    >>> remove_marks(order, latin_only=True)
    '“Herr Voß: • ½ cup of Œtker™ caffe latte • bowl of acai.”'

    >>> greek = 'Ζέφυρος, Zéfiro'
    >>> remove_marks(greek, latin_only=False)
    'Ζεφυρος, Zefiro'
    >>> remove_marks(greek, latin_only=True)
    'Ζέφυρος, Zefiro'

    Args:
        text:
            text to remove marks from
        latin_only:
            only remove marks from latin characters.

            "Often the reason to remove diacritics is to change Latin text to pure ASCII, but this
            function [when using `latin_only=False`] also changes non-Latin characters - like Greek
            letters - which will never become ASCII just by losing their accents. So it [often]
            makes sense to analyze each base character and to remove attached marks only if the
            base character is a letter from the Latin alphabet". (Fluent Python pg. 145-146).

    """
    if latin_only:
        norm_text = normalize('NFD', text)
        latin_base = False
        preserve = []
        for c in norm_text:
            if combining(c) and latin_base:
                continue
            preserve.append(c)

            if not combining(c):
                latin_base = c in string.ascii_letters
        shaved = ''.join(preserve)
    else:
        norm_text = normalize('NFD', text)
        shaved = ''.join(c for c in norm_text if not combining(c))

    return normalize('NFC', shaved)


_single_map = str.maketrans(
    """‚ƒ„ˆ‹‘’“”•–—˜›""",  # noqa: RUF001
    """'f"^<''""---~>""",
)
_multi_map = str.maketrans({
    '€': 'EUR',
    '…': '...',
    'Æ': 'AE',
    'æ': 'ae',
    'Œ': 'OE',
    'œ': 'oe',
    '™': '(TM)',
    '‰': '<per mille>',
    '†': '**',
    '‡': '***',
})
_multi_map.update(_single_map)


def _dewinize(text: str) -> str:
    """
    Replace Win1252 symbols with ASCII chars or sequences.

    This function is copied from:
        Fluent Python, 2nd ed., by Luciano Ramalho (O'Reilly). Pg. 146-147
        Copyright 2022 Luciano Ramalho, 978-1-492-05635-5
    """
    return text.translate(_multi_map)


def asciize(text: str) -> str:
    """
    This function:
        - removes all diacritic marks from latin base characters
        - replaces Win1252 symbols with ASCII chars
        - replaces common symbols in Westrn text (e.g., curly quotes, em dashes, bullets, etc.)
            into ASCII equivalents.

    This function and the examples below are copied/modified from:
        Fluent Python, 2nd ed., by Luciano Ramalho (O'Reilly). Pg. 147
        Copyright 2022 Luciano Ramalho, 978-1-492-05635-5

    >>> order = '“Herr Voß: • ½ cup of Œtker™ caffè latte • bowl of açaí.”'
    >>> _dewinize(order)
    '"Herr Voß: - ½ cup of OEtker(TM) caffè latte - bowl of açaí."'
    >>> asciize(order)
    '"Herr Voss: - 1⁄2 cup of OEtker(TM) caffe latte - bowl of acai."'

    Handling a string with Greek and Latin accented characters:

    >>> greek = 'Ζέφυρος, Zéfiro'
    >>> remove_marks(greek, latin_only=False)
    'Ζεφυρος, Zefiro'
    >>> remove_marks(greek, latin_only=True)
    'Ζέφυρος, Zefiro'
    >>> _dewinize(greek)
    'Ζέφυρος, Zéfiro'
    >>> asciize(greek)
    'Ζέφυρος, Zefiro'
    """  # noqa: D404, RUF002
    no_marks = remove_marks(_dewinize(text), latin_only=True)
    no_marks = no_marks.replace('ß', 'ss')
    return normalize('NFKC', no_marks)



import pytest
import pandas as pd


@pytest.fixture
def conversions():
    data = {
        'created_at': [
            '2023-01-01 00:00:00',  # Base date for the cohort
            '2023-01-01 01:00:00',  # Created an hour later than base date
            '2023-01-02 00:00:00',  # Created a day later
            '2023-01-15 00:00:00',  # Mid of the month, checking longer intervals
            '2023-01-25 00:00:00',  # Near the end of the month
            '2023-01-30 23:59:59',  # Edge case: Just before the end of the month
            '2023-01-01 00:00:01',  # Edge case: Created just after the base date
            '2023-01-31 00:00:00'   # Last day of the month
        ],
        'conversion_1': [
            '2023-01-01 00:00:01',  # Conversion happens in one second
            '2023-01-01 01:00:00',  # Conversion happens exactly at creation time (edge case)
            '2023-01-01 23:59:59',  # Conversion happens before the creation date
            '2023-01-16 00:00:00',  # Conversion happens the next day
            '2023-02-01 00:00:00',  # Conversion happens in the next month (checking longer intervals)  # noqa
            None,  # No conversion (edge case)
            '2023-01-01 00:00:00',  # Conversion happens before creation (logical error, edge case)
            '2023-01-31 00:01:00'   # Conversion happens shortly after creation
        ],
        'conversion_2': [
            '2023-01-01 00:01:00',  # Normal case, conversion after a minute
            '2023-01-01 01:01:00',  # Conversion happens after an hour
            '2023-01-02 00:01:00',  # Conversion happens after a minute
            '2023-02-01 00:00:00',  # Conversion happens in the next month
            '2023-02-24 00:00:00',  # Conversion nearly a month later
            '2023-01-31 23:59:59',  # Conversion at the very end of the month
            None,  # No conversion (edge case)
            '2023-01-31 00:01:00'   # Conversion happens shortly after creation
        ],
        'conversion_3': [
            '2023-01-01 00:00:02',  # Normal case, conversion in two seconds
            '2023-01-01 00:59:59',  # Conversion happens just before an hour completes
            '2023-01-03 00:00:00',  # Conversion after more than a day
            None,  # No conversion (edge case)
            '2023-02-24 23:59:59',  # Conversion nearly a month later, at the end of the day
            '2023-02-01 00:00:00',  # Conversion in the next month
            '2023-01-01 00:00:00',  # Conversion happens at the same time as creation (edge case)
            '2023-01-31 00:00:02'   # Conversion happens shortly after creation
        ],
        'segments': [
            'A',
            'B',
            'A',
            'B',
            'A',
            'A',
            'A',
            'A',
        ]
    }

    df = pd.DataFrame(data)
    df['created_at'] = pd.to_datetime(df['created_at'])
    df['conversion_1'] = pd.to_datetime(df['conversion_1'])
    df['conversion_2'] = pd.to_datetime(df['conversion_2'])
    df['conversion_3'] = pd.to_datetime(df['conversion_3'])
    return df



import unittest
import datetime
from enum import Enum, auto

import pandas as pd
import numpy as np

import helpsk.diff as diff
from helpsk import color
from tests.helpers import get_test_path


# noinspection PyMethodMayBeStatic
class TestDiff(unittest.TestCase):

    def test__create_html_change_span(self):
        self.assertEqual(
            diff._create_html_change_span(value='x', is_change=True),
            '<span style="background:#F1948A";>x</span>'
        )
        self.assertEqual(
            diff._create_html_change_span(value='x', is_change=False),
            '<span>x</span>'
        )

    def test__create_html_difference_list(self):
        self.assertEqual(
            diff._create_html_difference_list(value_a='TesT', value_b='test'),
            ['-T', '+t', ' e', ' s', '-T', '+t']
        )

    def test__create_html_cell(self):
        diff_list = diff._create_html_difference_list(value_a='TesT', value_b='test')
        self.assertEqual(
            diff._create_html_cell(difference_list=diff_list, is_first_value=True),
            '<span style="background:#F1948A";>T</span><span>e</span><span>s</span><span style="background:'  # noqa
            '#F1948A";>T</span>'
        )
        self.assertEqual(
            diff._create_html_cell(difference_list=diff_list, is_first_value=False),
            '<span style="background:#F1948A";>t</span><span>e</span><span>s</span><span style="background:'  # noqa
            '#F1948A";>t</span>'
        )

    def test__diff_dataframes(self):
        class TestEnum(Enum):
            VALUE_A = auto()
            VALUE_B = auto()

        df_a = pd.DataFrame({
            'col_1': [1, 1, 1, 1],
            'col_a': [np.nan, 2, 3, 4],
            'col_b': [np.nan, 'b', 'd', 'd'],
            'col_c': pd.date_range('2021-01-01', '2021-01-04'),
            'col_d': [None, np.nan, datetime.date(2021, 4, 2), datetime.date(2021, 4, 2)],
            'col_e': np.nan,
            'col_f': [1.0, 2.0, 3.0, 4.0],
            'col_h': [np.nan, TestEnum.VALUE_A, TestEnum.VALUE_B, TestEnum.VALUE_B],
            'col_i': [
                None,
                np.nan,
                datetime.datetime(2021, 4, 2, 0, 0, 0),
                datetime.datetime(2021, 4, 2, 0, 0, 0)
            ],
        })
        df_b = pd.DataFrame({
            'col_2': [2, 2, 2, 2],
            'col_a': [np.nan, 22, 13, 4],
            'col_b': [np.nan, 'b', 'ad', 'dz'],
            'col_c': pd.date_range('2022-01-02', '2022-01-05'),
            'col_d': [None, np.nan, datetime.date(2021, 5, 2), datetime.date(2021, 4, 6)],
            'col_e': np.nan,
            'col_f': [1.1, 22.0, 3.0, 4.0],
            'col_h': [np.nan, TestEnum.VALUE_B, TestEnum.VALUE_A, TestEnum.VALUE_B],
            'col_i': [
                None,
                np.nan,
                datetime.datetime(2021, 4, 2, 1, 1, 1),
                datetime.datetime(2021, 4, 2, 0, 0, 0)
            ],
        })
        html_a = diff.diff_dataframes(dataframe_a=df_a, dataframe_b=df_b)
        with open(get_test_path('diff/diff_dataframes.html'), 'w') as file:
            file.write(html_a)

        html_b = diff.diff_dataframes(dataframe_a=df_a, dataframe_b=df_b, change_color='#85C1E9')
        with open(get_test_path('diff/diff_dataframes_blue.html'), 'w') as file:
            file.write(html_b)

        # set index values to be different, shouldn't matter
        df_b.index = ['a', 'b', 'c', 'd']
        html_c = diff.diff_dataframes(dataframe_a=df_a, dataframe_b=df_b)
        self.assertEqual(html_a, html_c)

    def test__diff_test__strings(self):
        results = diff.diff_text(text_a='Hello', text_b='hello')
        with open(get_test_path('diff/diff_text__hello.html'), 'w') as file:
            file.write(results)

        results = diff.diff_text(
            text_a='Hello,\nthis is a message. Thanks! :)',
            text_b='hello, this is not a message. No Thanks :('
        )
        with open(get_test_path('diff/diff_text__new_line.html'), 'w') as file:
            file.write(results)

        results = diff.diff_text(
            text_a='Hello,\nthis is a message. Thanks! :)',
            text_b='hello, this is not a message. No Thanks :(',
            change_color=color.WARNING
        )
        with open(get_test_path('diff/diff_text__color.html'), 'w') as file:
            file.write(results)

    def test__diff_test__lists(self):
        results = diff.diff_text(
            text_a=[
                'hey',
                'Hello,\nthis is a message. Thanks! :)',
            ],
            text_b=[
                'Hello',
                'hello, this is not a message. No Thanks :(',
            ],
            change_color=color.WARNING
        )
        with open(get_test_path('diff/diff_text__lists.html'), 'w') as file:
            file.write(results)

    def test__diff_test__empty_strings(self):
        results = diff.diff_text(text_a='', text_b='')
        with open(get_test_path('diff/diff_text__empty_both.html'), 'w') as file:
            file.write(results)

        results = diff.diff_text(text_a='', text_b='Not Empty')
        with open(get_test_path('diff/diff_text__empty_a.html'), 'w') as file:
            file.write(results)

        results = diff.diff_text(text_a='Not Empty', text_b='')
        with open(get_test_path('diff/diff_text__empty_b.html'), 'w') as file:
            file.write(results)

        results = diff.diff_text(text_a=[], text_b=[])
        with open(get_test_path('diff/diff_text__empty_lists.html'), 'w') as file:
            file.write(results)

        results = diff.diff_text(text_a=[''], text_b=[''])
        with open(get_test_path('diff/diff_text__empty_list_values.html'), 'w') as file:
            file.write(results)

        results = diff.diff_text(text_a=[''], text_b=['Not Empty'])
        with open(get_test_path('diff/diff_text__empty_list_values_a.html'), 'w') as file:
            file.write(results)


if __name__ == '__main__':
    unittest.main()



"""Contains functions that are useful for logging."""

from __future__ import annotations
import logging
import datetime
from collections.abc import Callable
from functools import wraps


def log_function_call(max_param_chars: int = 100) -> Callable:
    """
    Decorator to log the function name and paramters of the function when called.

    Args:
        max_param_chars: the maximum number of characters to log for each param.
    """
    if isinstance(max_param_chars, int):
        def decorating_function(func: Callable) -> Callable:
            return _log_function_call_decorate(func=func, max_param_chars=max_param_chars)
        return decorating_function
    if callable(max_param_chars):
        # The func was passed in directly via the max_param_chars argument
        # This happens when the decorator is called without parentheses
        func, max_param_chars = max_param_chars, 100
        return _log_function_call_decorate(func=func, max_param_chars=max_param_chars)

    raise ValueError(f"Expected type of `int` or `callable`, received {type(max_param_chars)}")


def _log_function_call_decorate(func: Callable, max_param_chars: int) -> Callable:
    """
    Helper function that contains the actual decorator for `log_function_call`. It exists
    because log_function_call allows arguments and using the decorator without parentheses results
    in max_param_chars being set to the function we weant to wrap. This function allows
    log_function_call to the the correct parameters and pass them in to this function accordingly.
    I copied a similar pattern from the lru_cache decorator in
    https://github.com/python/cpython/blob/main/Lib/functools.py.
    """
    @wraps(func)
    def wrapper(*args: tuple, **kwargs: dict) -> Callable:
        func.__name__
        if len(args) == 0 and len(kwargs) == 0:
            _log_function(
                function_name=func.__name__,
                params=None,
                max_param_chars=max_param_chars,
            )
        else:
            parameters = {}
            if len(args) > 0:
                parameters['args'] = args
            if len(kwargs) > 0:
                parameters.update(kwargs)
            _log_function(
                function_name=func.__name__,
                params=parameters,
                max_param_chars=max_param_chars,
            )
        return func(*args, **kwargs)
    return wrapper


def _log_function(
        function_name: str,
        params: dict | None = None,
        max_param_chars: int = 100) -> None:
    """
    Meant to be used at the start of the calling function; calls logging.info and
    passes the name of the function and optional parameter names/values.

    Args:
        function_name:
            the name of the function
        params:
            a dictionary containing the names of the function parameters (as dictionary keys) and
            the parameter values (as dictionary values).
        max_param_chars:
            the maximum number of characters to log for each param.
    """
    logging.info(f"FUNCTION: {function_name.upper()}")
    if params is not None:
        logging.info("PARAMS:")
        for key, value in params.items():
            value = str(value)  # noqa: PLW2901
            if len(value) > max_param_chars:
                value = value[0:max_param_chars] + '...'  # noqa: PLW2901
            logging.info(f"    {key}: {value}")


class Timer:
    """Provides way to time the duration of code within the context manager."""

    def __init__(self, message: str, post_message: bool = False):
        """
        Args:
            message: message to show when timer starts
            post_message: if True, include same message when timer ends.
        """
        self._message = message
        self._post_message = post_message

    def __enter__(self):
        logging.basicConfig()
        logging.info(f'Timer Started: {self._message}')
        self._start = datetime.datetime.now()
        return self

    def __exit__(self, *args: tuple):
        self._end = datetime.datetime.now()
        self._interval = self._end - self._start
        message = ''
        if self._post_message:
            message = self._message + " "
        logging.info(f'Timer Finished: {message}({self._interval.total_seconds():.2f} seconds)')


def log_timer(function: Callable) -> Callable:
    """
    Function should be used as a decorator to log the time/duration of a function call.

    Args:
        function: function that is decorated
    """
    @wraps(function)
    def wrapper(*args: tuple, **kwargs: dict) -> Callable:
        with Timer(f"FUNCTION={function.__module__}:{function.__name__}", post_message=True):
            return function(*args, **kwargs)
    return wrapper



"""
Contains classes that wrap the connection/querying logic of various databases.
Database objects can be created from the __init__ function or from the .from_config class function
by passing in an instance of the corresponding configuration class.

Examples
    with Redshift(user='username', password='asdf', ....) as redshift:
        redshift.query("SELECT * FROM table LIMIT 100")

    or

    redshift_config = RedshiftConfigFile('/path/to/redshift.config')
    with Redshift.from_config(redshift_config) as redshift:
        redshift.query("SELECT * FROM table LIMIT 100")

"""
from __future__ import annotations
import pandas as pd
from helpsk.database_base import Database, ConnectionObject


class Redshift(Database):
    """
    Wraps logic for connecting to Redshift and querying.

    Example:
        Given redshift.config with contents:

            [redshift]
            dbname=my_username
            password=my-password-123
            database=the_database
            port=1234
            host=host.address.redshift.amazonaws.com

        with Redshift.from_config('/path/to/redshift.config', 'redshift') as redshift:
            redshift.query("SELECT * FROM table LIMIT 100")

        or

        with Redshift(dbname='my_username', password=...) as redshift:
            redshift.query("SELECT * FROM table LIMIT 100")
    """

    @property
    def _connection_string(self) -> str:
        return ' '.join(f"{k}={v}" for k, v in self._kwargs.items())

    def _open_connection_object(self) -> ConnectionObject:
        """Wraps logic for connecting to redshift."""
        from psycopg2 import connect
        return connect(self._connection_string)

    def _close_connection_object(self) -> None:
        """Wraps logic for closing the connection to redshift."""
        self.connection_object.close()

    def _query(self, sql: str) -> pd.DataFrame:
        """
        Wraps logic for querying redshift.

        Args:
            sql:
                SQL to execute e.g. "SELECT * FROM table LIMIT 100"

        Returns:
            a pandas Dataframe with the results from the query
        """
        return pd.read_sql_query(sql, self.connection_object)

    def execute_statement(self, statement: str) -> None:
        """Executes a statement without any data returned."""
        raise NotImplementedError()

    def insert_records(
            self,
            dataframe: pd.DataFrame, table: str, create_table: bool = False,
            overwrite: bool = True, schema: str | None = None,
            database: str | None = None) -> None:
        """Inserts rows into a table from a pandas DataFrame."""
        raise NotImplementedError()


class Snowflake(Database):
    """
    Wraps logic for connecting to Snowflake and querying.

    Example:
        config = SnowflakeConfigFile('/path/to/snowflake.config')
        with Snowflake.from_config(config) as snowflake:
            snowflake.query("SELECT * FROM table LIMIT 100")

    Instructions for installing snowflake dependencies:
        https://docs.snowflake.com/en/user-guide/python-connector-install.html

    I used v2.7.4, and python v3.9 (requirements_39):

    This page shows the latest versions:
        https://pypi.org/project/snowflake-connector-python/

    ```
    pip install -r .../snowflake-connector-python/v2.7.4/tested_requirements/requirements_39.reqs
    pip install snowflake-connector-python==v2.7.4
    ```

    Additionally:
        https://docs.snowflake.com/en/user-guide/python-connector-pandas.html#installation

    ```
    pip install snowflake-connector-python[pandas]
    ```
    """

    def __init__(self, **kwargs: dict):
        """Initialization."""
        super().__init__(**kwargs)

        if (autocommit := kwargs.get('autocommit')) is not None and isinstance(autocommit, str):
            if autocommit.lower() == 'false':
                self._kwargs['autocommit'] = False
            elif autocommit.lower() == 'true':
                self._kwargs['autocommit'] = True
            else:
                raise ValueError(f"autocommit needs to be true/false but received `{autocommit}`")

    def _open_connection_object(self) -> ConnectionObject:
        """Wraps logic for connecting to snowflake."""
        from snowflake.connector import connect
        return connect(**self._kwargs)

    def _close_connection_object(self) -> None:
        """Wraps logic for closing the connection to snowflake."""
        self.connection_object.close()

    def _query(self, sql: str) -> pd.DataFrame:
        """
        Wraps logic for querying snowflake.

        Args:
            sql:
                SQL to execute e.g. "SELECT * FROM table LIMIT 100"

        Returns:
            a pandas Dataframe with the results from the query
        """
        cursor = self.connection_object.cursor()
        cursor.execute(sql)
        dataframe = cursor.fetch_pandas_all()
        # We need to reset_index because there seems to be a bug in the connector that returns
        # duplicated index values. This can cause unexpected behavior downstream.
        # https://github.com/snowflakedb/snowflake-connector-python/issues/1061
        # https://stackoverflow.com/questions/69911999/none-unique-pandas-dataframe-index-created-using-cur-fetch-pandas-all-after-lo
        return dataframe.reset_index(drop=True)

    def execute_statement(self, statement: str) -> None:
        """Executes a statement without any data returned."""
        cursor = self.connection_object.cursor()
        results = cursor.execute(statement)
        return results.fetchall()

    @staticmethod
    def _generate_sql_create_table(
            table: str, dataframe: pd.DataFrame,
            database: str | None = None, schema: str | None = None) -> None:
        # adapted from https://stephenallwright.com/create-snowflake-table-pandas-dataframe/
        final_table_name = ''
        if database:
            final_table_name += f'{database.upper()}.'
        if schema:
            final_table_name += f'{schema.upper()}.'
        final_table_name += table.upper()

        create_statement = f"CREATE OR REPLACE TABLE {final_table_name} (\n"
        # Loop through each column finding the datatype and adding it to the statement
        for column in dataframe.columns:
            if dataframe[column].dtype.name in ('int', 'int64'):
                create_statement += f"    {column} int"
            elif dataframe[column].dtype.name == 'object':
                create_statement += f"    {column} varchar"
            elif dataframe[column].dtype.name == 'datetime64[ns]':
                create_statement += f"    {column} datetime"
            elif dataframe[column].dtype.name == 'float64':
                create_statement += f"    {column} float8"
            elif dataframe[column].dtype.name == 'bool':
                create_statement += f"    {column} boolean"
            else:
                create_statement += f"    {column} varchar"

            # If column is not last column, add comma, else end sql-query
            if dataframe[column].name != dataframe.columns[-1]:
                create_statement += ",\n"

        create_statement += "\n)"
        return create_statement

    def insert_records(self,
                       dataframe: pd.DataFrame,
                       table: str,
                       create_table: bool = False,
                       overwrite: bool = False,
                       schema: str | None = None,
                       database: str | None = None) -> None:
        """
        Inserts rows into a table from a pandas DataFrame.

        Args:
            dataframe:
                the pandas dataframe to insert
            table:
                the name of the table
            create_table:
                if True, creates the table before inserting
            overwrite:
                if True, drops all records before inserting
            schema:
                the name of the schema
            database:
                the name of the database
        """
        from snowflake.connector.pandas_tools import write_pandas
        dataframe = dataframe.copy()
        # snowflake is case sensitive and converts everything to upper-case
        dataframe.columns = dataframe.columns.str.upper()

        if create_table:
            create_table_sql = Snowflake._generate_sql_create_table(
                table=table,
                dataframe=dataframe,
                database=database,
                schema=schema,
            )
            _ = self.execute_statement(statement=create_table_sql)

        # if we are creating the table there is nothing to overwrite
        if not create_table and overwrite:
            final_table_name = ''
            if database:
                final_table_name += f'{database.upper()}.'
            if schema:
                final_table_name += f'{schema.upper()}.'
            final_table_name += table.upper()

            _ = self.execute_statement(f"DELETE FROM {final_table_name}")

        if database:
            database = database.upper()

        if schema:
            schema = schema.upper()

        _ = self.execute_statement(statement=f"USE SCHEMA {schema};")
        success, _, num_rows, _ = write_pandas(
            conn=self.connection_object,
            df=dataframe,
            database=database,
            schema=schema,
            table_name=table.upper(),
        )
        return success, num_rows


class Sqlite(Database):
    """
    Wraps logic for connecting to Redshift and querying.

        with Sqlite(path='/path/to/sqlite.db') as db:
            db.query("SELECT * FROM table LIMIT 100")
    """

    @property
    def _connection_string(self) -> str:
        return f"sqlite:///{self._kwargs['path']}"

    def _open_connection_object(self) -> ConnectionObject:
        """Wraps logic for connecting to sqlite."""
        from sqlalchemy import create_engine
        engine = create_engine(self._connection_string)
        return engine.connect()

    def _close_connection_object(self) -> None:
        """Wraps logic for closing the connection to sqlite."""
        self.connection_object.close()

    def _query(self, sql: str) -> pd.DataFrame:
        """
        Wraps logic for querying redshift.

        Args:
            sql:
                SQL to execute e.g. "SELECT * FROM table LIMIT 100"

        Returns:
            a pandas Dataframe with the results from the query
        """
        return pd.read_sql(sql, self.connection_object)

    def execute_statement(self, statement: str) -> None:
        """Executes a statement without any data returned."""
        self.connection_object.execute(statement)

    def insert_records(
            self,
            dataframe: pd.DataFrame,
            table: str,
            create_table: bool = False,
            overwrite: bool = False,
            schema: str | None = None,
            database: str | None = None) -> None:  # noqa: ARG002
        """
        Inserts rows into a table from a pandas DataFrame.

        Args:
            dataframe:
                the pandas dataframe to insert
            table:
                the name of the table
            create_table:
                if True, creates the table before inserting
            overwrite:
                if True, drops all records before inserting
            schema:
                the name of the schema
            database:
                Not used.
        """
        # if_exists{'fail', 'replace', 'append'}, default 'fail'
        #     How to behave if the table already exists.
        #     fail: Raise a ValueError.
        #     replace: Drop the table before inserting new values.
        #     append: Insert new values to the existing table.
        if create_table:
            if_exists = 'fail'
        elif overwrite:
            if_exists = 'replace'
        else:
            if_exists = 'append'

        _ = dataframe.to_sql(
            name=table,
            con=self.connection_object,
            schema=schema,
            if_exists=if_exists,
            index=False,
        )



"""Contains functions to help with comparing two strings or two dataframes."""

import difflib
from typing import Union
import pandas as pd


def _create_html_difference_list(value_a: str, value_b: str) -> str:
    """
    Given value_a and value_b return unified_diff, after cleaning up what unified_diff returns.

    Args:
        value_a: string to compare
        value_b: another string to compare
    """
    diff = list(difflib.unified_diff(a=value_a, b=value_b, n=1000))
    return [" " + char for char in value_a] if len(diff) == 0 else diff[3:]


def _create_html_change_span(value: str, is_change: bool, change_color: str = '#F1948A') -> str:
    """
    Creates a single character from one company's differences.

    Args:
        value: a single character
        is_change: if True, highlight the character in red
        change_color: color of background to highlight differences.

    Returns:
        e.g. "<span style="background:#ffe6e6";>value</span>"
    """
    background_color = ''
    if is_change:
        background_color = f' style="background:{change_color}";'
    return f'<span{background_color}>{value}</span>'


def _create_html_cell(
        difference_list: list,
        is_first_value: bool,
        change_color: str = '#F1948A') -> str:
    """
    Creates a single cell (e.g. name, domain, etc.) from one company's differences.

    Args:
        difference_list: list returned from create_difference_list
        is_first_value: if True, treats difference_list according to first value.
        change_color: color of background to highlight differences.
    """
    if is_first_value:
        diff = [(x[1:], x[0] != ' ') for x in difference_list if x[0] in [' ', '-']]
    else:
        diff = [(x[1:], x[0] != ' ') for x in difference_list if x[0] in [' ', '+']]

    html = [
        _create_html_change_span(value=x[0], is_change=x[1], change_color=change_color)
        for x in diff
    ]
    return ''.join(html)


def diff_text(
        text_a: Union[str, list[str]],
        text_b: Union[str, list[str]],
        change_color: str = '#F1948A') -> str:
    """
    Returns string as HTML containing highlighted differences between `text_a` and
    `text_b`.

    The HTML will contain a table with a single column that that contains `text_a` on top and
    `text_b` on the bottom.

    All `new line` characters are removed and replaced with a space.

    Args:
        text_a: this text will be represented on the top of each html cell.
        text_b: this text will be represented on the bottom of each html cell.
        change_color: color of background to highlight differences.
    """
    html = '<html><head><style> table, th, td { border: 1px solid black; border-collapse: ' \
        'collapse; white-space: normal;} </style></head><body style="font-family: monospace">'
    html += '<table><tr>'
    html += '<th>index</th>'
    html += '<th>diff</th>'
    html += '</tr>'

    line_break = '<hr style="width:25%; text-align:left; margin-left:10px; height:1px; ' \
        'border-width:0;' \
        'color:blue; background-color:blue">'

    def create_inline_change(diff_list: list[str]) -> str:
        diff_a = _create_html_cell(
            difference_list=diff_list, is_first_value=True, change_color=change_color,
        )
        diff_b = _create_html_cell(
            difference_list=diff_list, is_first_value=False, change_color=change_color,
        )
        return f"<td>{diff_a}<br>{line_break}{diff_b}</td>"

    if isinstance(text_a, str):
        assert isinstance(text_b, str)
        text_a = [text_a]
        text_b = [text_b]
    else:
        assert len(text_a) == len(text_b)

    for index in range(len(text_a)):
        html += '<tr>'
        html += f"<td>{index}</td>"
        difference_list = _create_html_difference_list(
            value_a=text_a[index],
            value_b=text_b[index],
        )
        html += create_inline_change(diff_list=difference_list)
        html += '</tr>'
    html += "</table></body></html>"
    return html


def diff_dataframes(
        dataframe_a: pd.DataFrame,
        dataframe_b: pd.DataFrame,
        change_color: str = '#F1948A') -> str:
    """
    Returns string as HTML containing highlighted differences between `dataframe_a` and
    `dataframe_b`.

    The HTML will contain a table showing columns that are found in both `dataframe_a` and
    `dataframe_b`.

    The DataFrames should be equal in length.

    Args:
        dataframe_a: this dataframe will be represented with values on the top of each html cell.
        dataframe_b: this dataframe will be represented with values on the bottom of each html
            cell.
        change_color: color of background to highlight differences.
    """
    assert len(dataframe_a) == len(dataframe_b)
    joint_columns = [x for x in dataframe_a.columns if x in dataframe_b.columns]
    assert len(joint_columns) > 0

    html = '<html><head><style> table, th, td { border: 1px solid black; border-collapse: ' \
        'collapse; white-space: normal;} </style></head><body style="font-family: monospace">'
    html += '<table><tr>'
    html += '<th>index</th>'
    for column in joint_columns:
        html += f'<th>{column}</th>'
    html += '</tr>'

    line_break = '<hr style="width:25%; text-align:left; margin-left:10px; height:1px; ' \
        'border-width:0;' \
        'color:blue; background-color:blue">'

    def create_inline_change(diff_list: list[str]) -> str:
        diff_a = _create_html_cell(
            difference_list=diff_list, is_first_value=True, change_color=change_color,
        )
        diff_b = _create_html_cell(
            difference_list=diff_list, is_first_value=False, change_color=change_color,
        )
        return f"<td>{diff_a}<br>{line_break}{diff_b}</td>"

    for index in range(len(dataframe_a)):
        html += '<tr>'
        html += f"<td>{index}</td>"
        for column in joint_columns:
            # print(f"{index}:{column}")
            value_a = str(dataframe_a[column].iloc[index])
            value_b = str(dataframe_b[column].iloc[index])
            difference_list = _create_html_difference_list(value_a=value_a, value_b=value_b)
            html += create_inline_change(diff_list=difference_list)

        html += '</tr>'
    html += "</table></body></html>"
    return html



import unittest

from helpsk.text import normalize_equal, remove_marks, asciize


# noinspection PyMethodMayBeStatic
class TestValidation(unittest.TestCase):

    def test_normalize_equal(self):
        self.assertTrue(normalize_equal('café', 'cafe\u0301'))
        self.assertTrue(normalize_equal('café', 'Cafe\u0301'))
        self.assertFalse(normalize_equal('café', 'Cafe\u0301', case_fold=False))
        ohm = '\u2126'
        omega = 'Ω'
        self.assertNotEqual(ohm, omega)
        self.assertTrue(normalize_equal(ohm, omega))
        self.assertTrue(normalize_equal(ohm, omega, case_fold=False))
        self.assertFalse(normalize_equal('Straße', 'strasse', case_fold=False))
        self.assertTrue(normalize_equal('Straße', 'strasse'))
        self.assertTrue(normalize_equal('a', 'A'))
        self.assertFalse(normalize_equal('a', 'A', case_fold=False))

    def test_remove_marks(self):
        order = '“Herr Voß: • ½ cup of Œtker™ caffè latte • bowl of açaí.”'
        expected = '“Herr Voß: • ½ cup of Œtker™ caffe latte • bowl of acai.”'
        self.assertEqual(expected, remove_marks(order, latin_only=False))
        self.assertEqual(expected, remove_marks(order, latin_only=True))
        greek = 'Ζέφυρος, Zéfiro'
        self.assertEqual('Ζεφυρος, Zefiro', remove_marks(greek, latin_only=False))
        self.assertEqual('Ζέφυρος, Zefiro', remove_marks(greek, latin_only=True))

    def test_asciize(self):
        order = '“Herr Voß: • ½ cup of Œtker™ caffè latte • bowl of açaí.”'
        self.assertEqual('"Herr Voss: - 1⁄2 cup of OEtker(TM) caffe latte - bowl of acai."', asciize(order))
        greek = 'Ζέφυρος, Zéfiro'
        self.assertEqual('Ζέφυρος, Zefiro', asciize(greek))



import datetime
import unittest

import numpy as np
import pandas as pd
from dateutil.parser import parse

from helpsk import date, validation
from tests.helpers import subtests_expected_vs_actual


# noinspection PyMethodMayBeStatic
class TestDate(unittest.TestCase):

    def test_fiscal_quarter_date(self):
        date_values = [
            '2020-12-01', '2020-12-15', '2020-12-31',
            '2021-01-01', '2021-01-15', '2021-01-31',
            '2021-02-01', '2021-02-15', '2021-02-28',
            '2021-03-01', '2021-03-15', '2021-03-31',
            '2021-04-01', '2021-04-15', '2021-04-30',
            '2021-05-01', '2021-05-15', '2021-05-31',
            '2021-06-01', '2021-06-15', '2021-06-30',
            '2021-07-01', '2021-07-15', '2021-07-31',
            '2021-08-01', '2021-08-15', '2021-08-31',
            '2021-09-01', '2021-09-15', '2021-09-30',
            '2021-10-01', '2021-10-15', '2021-10-31',
            '2021-11-01', '2021-11-15', '2021-11-30',
            '2021-12-01', '2021-12-15', '2021-12-31',
            '2022-01-01', '2022-01-15', '2022-01-31'
        ]

        test_parameters = dict(include_year=True, fiscal_start=1)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [
            2020.4, 2020.4, 2020.4,
            2021.1, 2021.1, 2021.1,
            2021.1, 2021.1, 2021.1,
            2021.1, 2021.1, 2021.1,
            2021.2, 2021.2, 2021.2,
            2021.2, 2021.2, 2021.2,
            2021.2, 2021.2, 2021.2,
            2021.3, 2021.3, 2021.3,
            2021.3, 2021.3, 2021.3,
            2021.3, 2021.3, 2021.3,
            2021.4, 2021.4, 2021.4,
            2021.4, 2021.4, 2021.4,
            2021.4, 2021.4, 2021.4,
            2022.1, 2022.1, 2022.1
        ]

        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(include_year=False, fiscal_start=1)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [
            4, 4, 4,
            1, 1, 1,
            1, 1, 1,
            1, 1, 1,
            2, 2, 2,
            2, 2, 2,
            2, 2, 2,
            3, 3, 3,
            3, 3, 3,
            3, 3, 3,
            4, 4, 4,
            4, 4, 4,
            4, 4, 4,
            1, 1, 1
        ]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(include_year=True, fiscal_start=2)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [
            2021.4, 2021.4, 2021.4,
            2021.4, 2021.4, 2021.4,
            2022.1, 2022.1, 2022.1,
            2022.1, 2022.1, 2022.1,
            2022.1, 2022.1, 2022.1,
            2022.2, 2022.2, 2022.2,
            2022.2, 2022.2, 2022.2,
            2022.2, 2022.2, 2022.2,
            2022.3, 2022.3, 2022.3,
            2022.3, 2022.3, 2022.3,
            2022.3, 2022.3, 2022.3,
            2022.4, 2022.4, 2022.4,
            2022.4, 2022.4, 2022.4,
            2022.4, 2022.4, 2022.4
        ]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(include_year=False, fiscal_start=2)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [
            4, 4, 4,
            4, 4, 4,
            1, 1, 1,
            1, 1, 1,
            1, 1, 1,
            2, 2, 2,
            2, 2, 2,
            2, 2, 2,
            3, 3, 3,
            3, 3, 3,
            3, 3, 3,
            4, 4, 4,
            4, 4, 4,
            4, 4, 4
        ]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(include_year=True, fiscal_start=12)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [
            2021.1, 2021.1, 2021.1,  # 2020-Dec
            2021.1, 2021.1, 2021.1,  # 2021-Jan
            2021.1, 2021.1, 2021.1,  # 2021-Feb
            2021.2, 2021.2, 2021.2,  # 2021-Mar
            2021.2, 2021.2, 2021.2,  # 2021-Apr
            2021.2, 2021.2, 2021.2,  # 2021-May
            2021.3, 2021.3, 2021.3,  # 2021-Jun
            2021.3, 2021.3, 2021.3,  # 2021-Jul
            2021.3, 2021.3, 2021.3,  # 2021-Aug
            2021.4, 2021.4, 2021.4,  # 2021-Sep
            2021.4, 2021.4, 2021.4,  # 2021-Oct
            2021.4, 2021.4, 2021.4,  # 2021-Nov
            2022.1, 2022.1, 2022.1,  # 2021-Dec
            2022.1, 2022.1, 2022.1,  # 2022-Jan
        ]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(include_year=False, fiscal_start=12)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [
            1, 1, 1,  # 2020-Dec
            1, 1, 1,  # 2021-Jan
            1, 1, 1,  # 2021-Feb
            2, 2, 2,  # 2021-Mar
            2, 2, 2,  # 2021-Apr
            2, 2, 2,  # 2021-May
            3, 3, 3,  # 2021-Jun
            3, 3, 3,  # 2021-Jul
            3, 3, 3,  # 2021-Aug
            4, 4, 4,  # 2021-Sep
            4, 4, 4,  # 2021-Oct
            4, 4, 4,  # 2021-Nov
            1, 1, 1,  # 2021-Dec
            1, 1, 1,  # 2022-Jan
        ]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

    def test_fiscal_quarter_datetime(self):
        date_values = [
            '2020-12-01', '2020-12-15', '2020-12-31',
            '2021-01-01', '2021-01-15', '2021-01-31',
            '2021-02-01', '2021-02-15', '2021-02-28',
            '2021-03-01', '2021-03-15', '2021-03-31',
            '2021-04-01', '2021-04-15', '2021-04-30',
            '2021-05-01', '2021-05-15', '2021-05-31',
            '2021-06-01', '2021-06-15', '2021-06-30',
            '2021-07-01', '2021-07-15', '2021-07-31',
            '2021-08-01', '2021-08-15', '2021-08-31',
            '2021-09-01', '2021-09-15', '2021-09-30',
            '2021-10-01', '2021-10-15', '2021-10-31',
            '2021-11-01', '2021-11-15', '2021-11-30',
            '2021-12-01', '2021-12-15', '2021-12-31',
            '2022-01-01', '2022-01-15', '2022-01-31',
        ]

        test_parameters = dict(include_year=True, fiscal_start=1)
        results = [date.fiscal_quarter(value=parse(x + ' 23:59:59'), **test_parameters)
                   for x in date_values]
        expected = [2020.4, 2020.4, 2020.4,
                    2021.1, 2021.1, 2021.1,
                    2021.1, 2021.1, 2021.1,
                    2021.1, 2021.1, 2021.1,
                    2021.2, 2021.2, 2021.2,
                    2021.2, 2021.2, 2021.2,
                    2021.2, 2021.2, 2021.2,
                    2021.3, 2021.3, 2021.3,
                    2021.3, 2021.3, 2021.3,
                    2021.3, 2021.3, 2021.3,
                    2021.4, 2021.4, 2021.4,
                    2021.4, 2021.4, 2021.4,
                    2021.4, 2021.4, 2021.4,
                    2022.1, 2022.1, 2022.1]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(include_year=False, fiscal_start=1)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [4, 4, 4,
                    1, 1, 1,
                    1, 1, 1,
                    1, 1, 1,
                    2, 2, 2,
                    2, 2, 2,
                    2, 2, 2,
                    3, 3, 3,
                    3, 3, 3,
                    3, 3, 3,
                    4, 4, 4,
                    4, 4, 4,
                    4, 4, 4,
                    1, 1, 1]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(include_year=True, fiscal_start=2)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [2021.4, 2021.4, 2021.4,
                    2021.4, 2021.4, 2021.4,
                    2022.1, 2022.1, 2022.1,
                    2022.1, 2022.1, 2022.1,
                    2022.1, 2022.1, 2022.1,
                    2022.2, 2022.2, 2022.2,
                    2022.2, 2022.2, 2022.2,
                    2022.2, 2022.2, 2022.2,
                    2022.3, 2022.3, 2022.3,
                    2022.3, 2022.3, 2022.3,
                    2022.3, 2022.3, 2022.3,
                    2022.4, 2022.4, 2022.4,
                    2022.4, 2022.4, 2022.4,
                    2022.4, 2022.4, 2022.4]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(include_year=False, fiscal_start=2)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [4, 4, 4,
                    4, 4, 4,
                    1, 1, 1,
                    1, 1, 1,
                    1, 1, 1,
                    2, 2, 2,
                    2, 2, 2,
                    2, 2, 2,
                    3, 3, 3,
                    3, 3, 3,
                    3, 3, 3,
                    4, 4, 4,
                    4, 4, 4,
                    4, 4, 4]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(include_year=True, fiscal_start=12)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [
            2021.1, 2021.1, 2021.1,  # 2020-Dec
            2021.1, 2021.1, 2021.1,  # 2021-Jan
            2021.1, 2021.1, 2021.1,  # 2021-Feb
            2021.2, 2021.2, 2021.2,  # 2021-Mar
            2021.2, 2021.2, 2021.2,  # 2021-Apr
            2021.2, 2021.2, 2021.2,  # 2021-May
            2021.3, 2021.3, 2021.3,  # 2021-Jun
            2021.3, 2021.3, 2021.3,  # 2021-Jul
            2021.3, 2021.3, 2021.3,  # 2021-Aug
            2021.4, 2021.4, 2021.4,  # 2021-Sep
            2021.4, 2021.4, 2021.4,  # 2021-Oct
            2021.4, 2021.4, 2021.4,  # 2021-Nov
            2022.1, 2022.1, 2022.1,  # 2021-Dec
            2022.1, 2022.1, 2022.1,  # 2022-Jan
        ]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(include_year=False, fiscal_start=12)
        results = [date.fiscal_quarter(value=parse(x), **test_parameters) for x in date_values]
        expected = [
            1, 1, 1,  # 2020-Dec
            1, 1, 1,  # 2021-Jan
            1, 1, 1,  # 2021-Feb
            2, 2, 2,  # 2021-Mar
            2, 2, 2,  # 2021-Apr
            2, 2, 2,  # 2021-May
            3, 3, 3,  # 2021-Jun
            3, 3, 3,  # 2021-Jul
            3, 3, 3,  # 2021-Aug
            4, 4, 4,  # 2021-Sep
            4, 4, 4,  # 2021-Oct
            4, 4, 4,  # 2021-Nov
            1, 1, 1,  # 2021-Dec
            1, 1, 1,  # 2022-Jan
        ]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

    def test_to_string_date(self):
        date_values = [
            '2020-12-01', '2020-12-15', '2020-12-31',
            '2021-01-01', '2021-01-15', '2021-01-31',
            '2021-02-01', '2021-02-15', '2021-02-28',
            '2021-03-01', '2021-03-15', '2021-03-31',
            '2021-04-01', '2021-04-15', '2021-04-30',
            '2021-05-01', '2021-05-15', '2021-05-31',
            '2021-06-01', '2021-06-15', '2021-06-30',
            '2021-07-01', '2021-07-15', '2021-07-31',
            '2021-08-01', '2021-08-15', '2021-08-31',
            '2021-09-01', '2021-09-15', '2021-09-30',
            '2021-10-01', '2021-10-15', '2021-10-31',
            '2021-11-01', '2021-11-15', '2021-11-30',
            '2021-12-01', '2021-12-15', '2021-12-31',
            '2022-01-01', '2022-01-15', '2022-01-31',
        ]

        test_parameters = dict(granularity=date.Granularity.DAY)
        results = [date.to_string(value=parse(x), **test_parameters) for x in date_values]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=date_values,
            **test_parameters
        )

        test_parameters = dict(granularity=date.Granularity.MONTH)
        results = [date.to_string(value=parse(x), **test_parameters) for x in date_values]
        expected = ['2020-Dec', '2020-Dec', '2020-Dec',
                    '2021-Jan', '2021-Jan', '2021-Jan',
                    '2021-Feb', '2021-Feb', '2021-Feb',
                    '2021-Mar', '2021-Mar', '2021-Mar',
                    '2021-Apr', '2021-Apr', '2021-Apr',
                    '2021-May', '2021-May', '2021-May',
                    '2021-Jun', '2021-Jun', '2021-Jun',
                    '2021-Jul', '2021-Jul', '2021-Jul',
                    '2021-Aug', '2021-Aug', '2021-Aug',
                    '2021-Sep', '2021-Sep', '2021-Sep',
                    '2021-Oct', '2021-Oct', '2021-Oct',
                    '2021-Nov', '2021-Nov', '2021-Nov',
                    '2021-Dec', '2021-Dec', '2021-Dec',
                    '2022-Jan', '2022-Jan', '2022-Jan']
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(granularity=date.Granularity.QUARTER, fiscal_start=1)
        results = [date.to_string(value=parse(x), **test_parameters) for x in date_values]
        expected = ['2020-Q4', '2020-Q4', '2020-Q4',  # 2020-Dec
                    '2021-Q1', '2021-Q1', '2021-Q1',  # 2021-Jan
                    '2021-Q1', '2021-Q1', '2021-Q1',  # 2021-Feb
                    '2021-Q1', '2021-Q1', '2021-Q1',  # 2021-Mar
                    '2021-Q2', '2021-Q2', '2021-Q2',  # 2021-Apr
                    '2021-Q2', '2021-Q2', '2021-Q2',  # 2021-May
                    '2021-Q2', '2021-Q2', '2021-Q2',  # 2021-Jun
                    '2021-Q3', '2021-Q3', '2021-Q3',  # 2021-Jul
                    '2021-Q3', '2021-Q3', '2021-Q3',  # 2021-Aug
                    '2021-Q3', '2021-Q3', '2021-Q3',  # 2021-Sep
                    '2021-Q4', '2021-Q4', '2021-Q4',  # 2021-Oct
                    '2021-Q4', '2021-Q4', '2021-Q4',  # 2021-Nov
                    '2021-Q4', '2021-Q4', '2021-Q4',  # 2021-Dec
                    '2022-Q1', '2022-Q1', '2022-Q1']  # 2022-Jan
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(granularity=date.Granularity.QUARTER, fiscal_start=2)
        results = [date.to_string(value=parse(x), **test_parameters) for x in date_values]
        expected = ['2021-FQ4', '2021-FQ4', '2021-FQ4',  # 2020-Dec
                    '2021-FQ4', '2021-FQ4', '2021-FQ4',  # 2021-Jan
                    '2022-FQ1', '2022-FQ1', '2022-FQ1',  # 2021-Feb
                    '2022-FQ1', '2022-FQ1', '2022-FQ1',  # 2021-Mar
                    '2022-FQ1', '2022-FQ1', '2022-FQ1',  # 2021-Apr
                    '2022-FQ2', '2022-FQ2', '2022-FQ2',  # 2021-May
                    '2022-FQ2', '2022-FQ2', '2022-FQ2',  # 2021-Jun
                    '2022-FQ2', '2022-FQ2', '2022-FQ2',  # 2021-Jul
                    '2022-FQ3', '2022-FQ3', '2022-FQ3',  # 2021-Aug
                    '2022-FQ3', '2022-FQ3', '2022-FQ3',  # 2021-Sep
                    '2022-FQ3', '2022-FQ3', '2022-FQ3',  # 2021-Oct
                    '2022-FQ4', '2022-FQ4', '2022-FQ4',  # 2021-Nov
                    '2022-FQ4', '2022-FQ4', '2022-FQ4',  # 2021-Dec
                    '2022-FQ4', '2022-FQ4', '2022-FQ4']  # 2022-Jan
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(granularity=date.Granularity.QUARTER, fiscal_start=12)
        results = [date.to_string(value=parse(x), **test_parameters) for x in date_values]
        expected = ['2021-FQ1', '2021-FQ1', '2021-FQ1',  # 2020-Dec
                    '2021-FQ1', '2021-FQ1', '2021-FQ1',  # 2021-Jan
                    '2021-FQ1', '2021-FQ1', '2021-FQ1',  # 2021-Feb
                    '2021-FQ2', '2021-FQ2', '2021-FQ2',  # 2021-Mar
                    '2021-FQ2', '2021-FQ2', '2021-FQ2',  # 2021-Apr
                    '2021-FQ2', '2021-FQ2', '2021-FQ2',  # 2021-May
                    '2021-FQ3', '2021-FQ3', '2021-FQ3',  # 2021-Jun
                    '2021-FQ3', '2021-FQ3', '2021-FQ3',  # 2021-Jul
                    '2021-FQ3', '2021-FQ3', '2021-FQ3',  # 2021-Aug
                    '2021-FQ4', '2021-FQ4', '2021-FQ4',  # 2021-Sep
                    '2021-FQ4', '2021-FQ4', '2021-FQ4',  # 2021-Oct
                    '2021-FQ4', '2021-FQ4', '2021-FQ4',  # 2021-Nov
                    '2022-FQ1', '2022-FQ1', '2022-FQ1',  # 2021-Dec
                    '2022-FQ1', '2022-FQ1', '2022-FQ1']  # 2022-Jan
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

    def test_to_string_datetime(self):
        date_values = [
            '2020-12-01', '2020-12-15', '2020-12-31',
            '2021-01-01', '2021-01-15', '2021-01-31',
            '2021-02-01', '2021-02-15', '2021-02-28',
            '2021-03-01', '2021-03-15', '2021-03-31',
            '2021-04-01', '2021-04-15', '2021-04-30',
            '2021-05-01', '2021-05-15', '2021-05-31',
            '2021-06-01', '2021-06-15', '2021-06-30',
            '2021-07-01', '2021-07-15', '2021-07-31',
            '2021-08-01', '2021-08-15', '2021-08-31',
            '2021-09-01', '2021-09-15', '2021-09-30',
            '2021-10-01', '2021-10-15', '2021-10-31',
            '2021-11-01', '2021-11-15', '2021-11-30',
            '2021-12-01', '2021-12-15', '2021-12-31',
            '2022-01-01', '2022-01-15', '2022-01-31',
        ]

        test_parameters = dict(granularity=date.Granularity.DAY)
        results = [date.to_string(value=parse(x + ' 23:59:59'), **test_parameters)
                   for x in date_values]
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=date_values,
            **test_parameters
        )

        test_parameters = dict(granularity=date.Granularity.MONTH)
        results = [date.to_string(value=parse(x), **test_parameters) for x in date_values]
        expected = ['2020-Dec', '2020-Dec', '2020-Dec',
                    '2021-Jan', '2021-Jan', '2021-Jan',
                    '2021-Feb', '2021-Feb', '2021-Feb',
                    '2021-Mar', '2021-Mar', '2021-Mar',
                    '2021-Apr', '2021-Apr', '2021-Apr',
                    '2021-May', '2021-May', '2021-May',
                    '2021-Jun', '2021-Jun', '2021-Jun',
                    '2021-Jul', '2021-Jul', '2021-Jul',
                    '2021-Aug', '2021-Aug', '2021-Aug',
                    '2021-Sep', '2021-Sep', '2021-Sep',
                    '2021-Oct', '2021-Oct', '2021-Oct',
                    '2021-Nov', '2021-Nov', '2021-Nov',
                    '2021-Dec', '2021-Dec', '2021-Dec',
                    '2022-Jan', '2022-Jan', '2022-Jan']
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(granularity=date.Granularity.QUARTER, fiscal_start=1)
        results = [date.to_string(value=parse(x), **test_parameters) for x in date_values]
        expected = ['2020-Q4', '2020-Q4', '2020-Q4',  # 2020-Dec
                    '2021-Q1', '2021-Q1', '2021-Q1',  # 2021-Jan
                    '2021-Q1', '2021-Q1', '2021-Q1',  # 2021-Feb
                    '2021-Q1', '2021-Q1', '2021-Q1',  # 2021-Mar
                    '2021-Q2', '2021-Q2', '2021-Q2',  # 2021-Apr
                    '2021-Q2', '2021-Q2', '2021-Q2',  # 2021-May
                    '2021-Q2', '2021-Q2', '2021-Q2',  # 2021-Jun
                    '2021-Q3', '2021-Q3', '2021-Q3',  # 2021-Jul
                    '2021-Q3', '2021-Q3', '2021-Q3',  # 2021-Aug
                    '2021-Q3', '2021-Q3', '2021-Q3',  # 2021-Sep
                    '2021-Q4', '2021-Q4', '2021-Q4',  # 2021-Oct
                    '2021-Q4', '2021-Q4', '2021-Q4',  # 2021-Nov
                    '2021-Q4', '2021-Q4', '2021-Q4',  # 2021-Dec
                    '2022-Q1', '2022-Q1', '2022-Q1']  # 2022-Jan
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(granularity=date.Granularity.QUARTER, fiscal_start=2)
        results = [date.to_string(value=parse(x), **test_parameters) for x in date_values]
        expected = ['2021-FQ4', '2021-FQ4', '2021-FQ4',  # 2020-Dec
                    '2021-FQ4', '2021-FQ4', '2021-FQ4',  # 2021-Jan
                    '2022-FQ1', '2022-FQ1', '2022-FQ1',  # 2021-Feb
                    '2022-FQ1', '2022-FQ1', '2022-FQ1',  # 2021-Mar
                    '2022-FQ1', '2022-FQ1', '2022-FQ1',  # 2021-Apr
                    '2022-FQ2', '2022-FQ2', '2022-FQ2',  # 2021-May
                    '2022-FQ2', '2022-FQ2', '2022-FQ2',  # 2021-Jun
                    '2022-FQ2', '2022-FQ2', '2022-FQ2',  # 2021-Jul
                    '2022-FQ3', '2022-FQ3', '2022-FQ3',  # 2021-Aug
                    '2022-FQ3', '2022-FQ3', '2022-FQ3',  # 2021-Sep
                    '2022-FQ3', '2022-FQ3', '2022-FQ3',  # 2021-Oct
                    '2022-FQ4', '2022-FQ4', '2022-FQ4',  # 2021-Nov
                    '2022-FQ4', '2022-FQ4', '2022-FQ4',  # 2021-Dec
                    '2022-FQ4', '2022-FQ4', '2022-FQ4']  # 2022-Jan
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

        test_parameters = dict(granularity=date.Granularity.QUARTER, fiscal_start=12)
        results = [date.to_string(value=parse(x), **test_parameters) for x in date_values]
        expected = ['2021-FQ1', '2021-FQ1', '2021-FQ1',  # 2020-Dec
                    '2021-FQ1', '2021-FQ1', '2021-FQ1',  # 2021-Jan
                    '2021-FQ1', '2021-FQ1', '2021-FQ1',  # 2021-Feb
                    '2021-FQ2', '2021-FQ2', '2021-FQ2',  # 2021-Mar
                    '2021-FQ2', '2021-FQ2', '2021-FQ2',  # 2021-Apr
                    '2021-FQ2', '2021-FQ2', '2021-FQ2',  # 2021-May
                    '2021-FQ3', '2021-FQ3', '2021-FQ3',  # 2021-Jun
                    '2021-FQ3', '2021-FQ3', '2021-FQ3',  # 2021-Jul
                    '2021-FQ3', '2021-FQ3', '2021-FQ3',  # 2021-Aug
                    '2021-FQ4', '2021-FQ4', '2021-FQ4',  # 2021-Sep
                    '2021-FQ4', '2021-FQ4', '2021-FQ4',  # 2021-Oct
                    '2021-FQ4', '2021-FQ4', '2021-FQ4',  # 2021-Nov
                    '2022-FQ1', '2022-FQ1', '2022-FQ1',  # 2021-Dec
                    '2022-FQ1', '2022-FQ1', '2022-FQ1']  # 2022-Jan
        subtests_expected_vs_actual(
            test_case=self, actual_values=results, expected_values=expected,
            **test_parameters
        )

    def test__to_string__handles_none_nat_empty(self):
        value = ''
        assert date.to_string(value, granularity=date.Granularity.DAY) is None
        assert date.to_string(value, granularity=date.Granularity.MONTH) is None
        assert date.to_string(value, granularity=date.Granularity.QUARTER) is None

        value = None
        assert date.to_string(value, granularity=date.Granularity.DAY) is None
        assert date.to_string(value, granularity=date.Granularity.MONTH) is None
        assert date.to_string(value, granularity=date.Granularity.QUARTER) is None

        value = pd.NaT
        assert date.to_string(value, granularity=date.Granularity.DAY) is None
        assert date.to_string(value, granularity=date.Granularity.MONTH) is None
        assert date.to_string(value, granularity=date.Granularity.QUARTER) is None

        value = np.NaN
        assert date.to_string(value, granularity=date.Granularity.DAY) is None
        assert date.to_string(value, granularity=date.Granularity.MONTH) is None
        assert date.to_string(value, granularity=date.Granularity.QUARTER) is None

        date_values = ['2020-12-01', '2020-12-15', None, pd.NaT]
        date_series = pd.to_datetime(pd.Series(date_values))

        expected_values = ['2020-12-01', '2020-12-15', pd.NaT, pd.NaT]
        actual_values = date_series.apply(
            lambda x: date.to_string(x, granularity=date.Granularity.DAY)
        )
        assert actual_values.iloc[0] == expected_values[0]
        assert actual_values.iloc[1] == expected_values[1]
        assert actual_values.iloc[2] is None
        assert actual_values.iloc[3] is None

        expected_values = ['2020-Dec', '2020-Dec', pd.NaT, pd.NaT]
        actual_values = date_series.apply(
            lambda x: date.to_string(x, granularity=date.Granularity.MONTH)
        )
        assert actual_values.iloc[0] == expected_values[0]
        assert actual_values.iloc[1] == expected_values[1]
        assert actual_values.iloc[2] is None
        assert actual_values.iloc[3] is None

        expected_values = ['2020-Q4', '2020-Q4', pd.NaT, pd.NaT]
        actual_values = date_series.apply(
            lambda x: date.to_string(x, granularity=date.Granularity.QUARTER)
        )
        assert actual_values.iloc[0] == expected_values[0]
        assert actual_values.iloc[1] == expected_values[1]
        assert actual_values.iloc[2] is None
        assert actual_values.iloc[3] is None

        expected_values = ['2021-FQ4', '2021-FQ4', pd.NaT, pd.NaT]
        actual_values = date_series.apply(
            lambda x: date.to_string(x, granularity=date.Granularity.QUARTER, fiscal_start=2)
        )
        assert actual_values.iloc[0] == expected_values[0]
        assert actual_values.iloc[1] == expected_values[1]
        assert actual_values.iloc[2] is None
        assert actual_values.iloc[3] is None

    def test_floor_missing_value(self):
        self.assertTrue(date.floor(pd.NA, granularity=date.Granularity.DAY) is pd.NA)
        self.assertTrue(date.floor(pd.NaT, granularity=date.Granularity.DAY) is pd.NaT)
        self.assertTrue(date.floor(np.NaN, granularity=date.Granularity.DAY) is np.NaN)
        self.assertTrue(date.floor(None, granularity=date.Granularity.DAY) is None)
        self.assertTrue(date.floor(pd.NA, granularity=date.Granularity.MONTH) is pd.NA)
        self.assertTrue(date.floor(pd.NaT, granularity=date.Granularity.MONTH) is pd.NaT)
        self.assertTrue(date.floor(np.NaN, granularity=date.Granularity.MONTH) is np.NaN)
        self.assertTrue(date.floor(None, granularity=date.Granularity.MONTH) is None)
        self.assertTrue(date.floor(pd.NA, granularity=date.Granularity.QUARTER) is pd.NA)
        self.assertTrue(date.floor(pd.NaT, granularity=date.Granularity.QUARTER) is pd.NaT)
        self.assertTrue(date.floor(np.NaN, granularity=date.Granularity.QUARTER) is np.NaN)
        self.assertTrue(date.floor(None, granularity=date.Granularity.QUARTER) is None)

    def test_floor_day(self):
        # test datetime
        value = datetime.datetime(year=2021, month=2, day=13, hour=23, minute=45, second=55)
        self.assertEqual(date.floor(value, granularity=date.Granularity.DAY),
                         parse('2021-02-13').date())
        self.assertEqual(date.floor(value),
                         parse('2021-02-13').date())
        # test date
        value = datetime.date(year=2021, month=2, day=13)
        self.assertEqual(date.floor(value, granularity=date.Granularity.DAY),
                         parse('2021-02-13').date())
        self.assertEqual(date.floor(value),
                         parse('2021-02-13').date())

    def test_floor_month(self):
        # test datetime
        value = datetime.datetime(year=2021, month=1, day=1, hour=23, minute=45, second=55)
        self.assertEqual(date.floor(value, granularity=date.Granularity.MONTH),
                         parse('2021-01-01').date())
        value = datetime.datetime(year=2021, month=1, day=31, hour=23, minute=45, second=55)
        self.assertEqual(date.floor(value, granularity=date.Granularity.MONTH),
                         parse('2021-01-01').date())
        value = datetime.datetime(year=2021, month=12, day=1, hour=23, minute=45, second=55)
        self.assertEqual(date.floor(value, granularity=date.Granularity.MONTH),
                         parse('2021-12-01').date())
        value = datetime.datetime(year=2021, month=12, day=31, hour=23, minute=45, second=55)
        self.assertEqual(date.floor(value, granularity=date.Granularity.MONTH),
                         parse('2021-12-01').date())

        # test date
        self.assertEqual(date.floor(parse('2021-01-01'), granularity=date.Granularity.MONTH),
                         parse('2021-01-01').date())
        self.assertEqual(date.floor(parse('2021-01-31'), granularity=date.Granularity.MONTH),
                         parse('2021-01-01').date())
        self.assertEqual(date.floor(parse('2021-12-01'), granularity=date.Granularity.MONTH),
                         parse('2021-12-01').date())
        self.assertEqual(date.floor(parse('2021-12-31'), granularity=date.Granularity.MONTH),
                         parse('2021-12-01').date())

    def test_floor_quarter(self):
        # default argument fiscal_start of 1
        self.assertEqual(date.floor(parse('2021-01-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-01-01').date())
        self.assertEqual(date.floor(parse('2021-01-31'), granularity=date.Granularity.QUARTER),
                         parse('2021-01-01').date())
        self.assertEqual(date.floor(parse('2021-02-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-01-01').date())
        self.assertEqual(date.floor(parse('2021-02-28'), granularity=date.Granularity.QUARTER),
                         parse('2021-01-01').date())
        self.assertEqual(date.floor(parse('2021-03-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-01-01').date())
        self.assertEqual(date.floor(parse('2021-03-31'), granularity=date.Granularity.QUARTER),
                         parse('2021-01-01').date())
        self.assertEqual(date.floor(parse('2021-04-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-04-01').date())
        self.assertEqual(date.floor(parse('2021-04-30'), granularity=date.Granularity.QUARTER),
                         parse('2021-04-01').date())
        self.assertEqual(date.floor(parse('2021-05-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-04-01').date())
        self.assertEqual(date.floor(parse('2021-05-31'), granularity=date.Granularity.QUARTER),
                         parse('2021-04-01').date())
        self.assertEqual(date.floor(parse('2021-06-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-04-01').date())
        self.assertEqual(date.floor(parse('2021-06-30'), granularity=date.Granularity.QUARTER),
                         parse('2021-04-01').date())
        self.assertEqual(date.floor(parse('2021-07-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-07-01').date())
        self.assertEqual(date.floor(parse('2021-07-31'), granularity=date.Granularity.QUARTER),
                         parse('2021-07-01').date())
        self.assertEqual(date.floor(parse('2021-08-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-07-01').date())
        self.assertEqual(date.floor(parse('2021-08-31'), granularity=date.Granularity.QUARTER),
                         parse('2021-07-01').date())
        self.assertEqual(date.floor(parse('2021-09-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-07-01').date())
        self.assertEqual(date.floor(parse('2021-09-30'), granularity=date.Granularity.QUARTER),
                         parse('2021-07-01').date())
        self.assertEqual(date.floor(parse('2021-10-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-10-01').date())
        self.assertEqual(date.floor(parse('2021-10-31'), granularity=date.Granularity.QUARTER),
                         parse('2021-10-01').date())
        self.assertEqual(date.floor(parse('2021-11-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-10-01').date())
        self.assertEqual(date.floor(parse('2021-11-30'), granularity=date.Granularity.QUARTER),
                         parse('2021-10-01').date())
        self.assertEqual(date.floor(parse('2021-12-01'), granularity=date.Granularity.QUARTER),
                         parse('2021-10-01').date())
        self.assertEqual(date.floor(parse('2021-12-31'), granularity=date.Granularity.QUARTER),
                         parse('2021-10-01').date())

        # fiscal quarter starts in February
        self.assertEqual(date.floor(parse('2021-01-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2020-11-01').date())
        self.assertEqual(date.floor(parse('2021-01-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2020-11-01').date())
        self.assertEqual(date.floor(parse('2021-02-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-02-28'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-03-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-03-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-04-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-04-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-05-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-05-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-06-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-06-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-07-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-07-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-08-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-08-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-09-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-09-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-10-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-10-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-11-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-11-01').date())
        self.assertEqual(date.floor(parse('2021-11-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-11-01').date())
        self.assertEqual(date.floor(parse('2021-12-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-11-01').date())
        self.assertEqual(date.floor(parse('2021-12-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=2),
                         parse('2021-11-01').date())

        # fiscal quarter starts in November (should be same as February)
        self.assertEqual(date.floor(parse('2021-01-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2020-11-01').date())
        self.assertEqual(date.floor(parse('2021-01-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2020-11-01').date())
        self.assertEqual(date.floor(parse('2021-02-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-02-28'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-03-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-03-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-04-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-04-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-02-01').date())
        self.assertEqual(date.floor(parse('2021-05-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-05-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-06-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-06-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-07-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-07-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-05-01').date())
        self.assertEqual(date.floor(parse('2021-08-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-08-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-09-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-09-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-10-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-10-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-08-01').date())
        self.assertEqual(date.floor(parse('2021-11-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-11-01').date())
        self.assertEqual(date.floor(parse('2021-11-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-11-01').date())
        self.assertEqual(date.floor(parse('2021-12-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-11-01').date())
        self.assertEqual(date.floor(parse('2021-12-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=11),
                         parse('2021-11-01').date())

        # fiscal quarter starts in June
        self.assertEqual(date.floor(parse('2021-01-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2020-12-01').date())
        self.assertEqual(date.floor(parse('2021-01-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2020-12-01').date())
        self.assertEqual(date.floor(parse('2021-02-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2020-12-01').date())
        self.assertEqual(date.floor(parse('2021-02-28'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2020-12-01').date())
        self.assertEqual(date.floor(parse('2021-03-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-03-01').date())
        self.assertEqual(date.floor(parse('2021-03-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-03-01').date())
        self.assertEqual(date.floor(parse('2021-04-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-03-01').date())
        self.assertEqual(date.floor(parse('2021-04-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-03-01').date())
        self.assertEqual(date.floor(parse('2021-05-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-03-01').date())
        self.assertEqual(date.floor(parse('2021-05-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-03-01').date())
        self.assertEqual(date.floor(parse('2021-06-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-06-01').date())
        self.assertEqual(date.floor(parse('2021-06-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-06-01').date())
        self.assertEqual(date.floor(parse('2021-07-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-06-01').date())
        self.assertEqual(date.floor(parse('2021-07-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-06-01').date())
        self.assertEqual(date.floor(parse('2021-08-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-06-01').date())
        self.assertEqual(date.floor(parse('2021-08-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-06-01').date())
        self.assertEqual(date.floor(parse('2021-09-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-09-01').date())
        self.assertEqual(date.floor(parse('2021-09-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-09-01').date())
        self.assertEqual(date.floor(parse('2021-10-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-09-01').date())
        self.assertEqual(date.floor(parse('2021-10-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-09-01').date())
        self.assertEqual(date.floor(parse('2021-11-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-09-01').date())
        self.assertEqual(date.floor(parse('2021-11-30'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-09-01').date())
        self.assertEqual(date.floor(parse('2021-12-01'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-12-01').date())
        self.assertEqual(date.floor(parse('2021-12-31'), granularity=date.Granularity.QUARTER,
                                    fiscal_start=6),
                         parse('2021-12-01').date())

    def test_floor_series(self):
        datetimes = pd.to_datetime([
            '2021-01-01 00:00:00', '2021-01-01 00:00:01',
            np.NaN,
            '2021-01-02 00:00:01', '2021-01-02 23:59:59',
            '2021-02-01 00:00:00', '2021-02-01 00:00:01',
            np.NaN,
            '2021-02-02 00:00:01', '2021-02-02 23:59:59',
            '2021-03-01 00:00:00', '2021-03-01 00:00:01',
            np.NaN,
            '2021-03-02 00:00:01', '2021-03-02 23:59:59',
            '2021-04-01 00:00:00', '2021-04-01 00:00:01',
            np.NaN,
            '2021-04-02 00:00:01', '2021-04-02 23:59:59',
            '2021-05-01 00:00:00', '2021-05-01 00:00:01',
            np.NaN,
            '2021-05-02 00:00:01', '2021-05-02 23:59:59',
            '2021-06-01 00:00:00', '2021-06-01 00:00:01',
            np.NaN,
            '2021-06-02 00:00:01', '2021-06-02 23:59:59',
            '2021-07-01 00:00:00', '2021-07-01 00:00:01',
            np.NaN,
            '2021-07-02 00:00:01', '2021-07-02 23:59:59',
            '2021-08-01 00:00:00', '2021-08-01 00:00:01',
            np.NaN,
            '2021-08-02 00:00:01', '2021-08-02 23:59:59',
            '2021-09-01 00:00:00', '2021-09-01 00:00:01',
            np.NaN,
            '2021-09-02 00:00:01', '2021-09-02 23:59:59',
            '2021-10-01 00:00:00', '2021-10-01 00:00:01',
            np.NaN,
            '2021-10-02 00:00:01', '2021-10-02 23:59:59',
            '2021-11-01 00:00:00', '2021-11-01 00:00:01',
            np.NaN,
            '2021-11-02 00:00:01', '2021-11-02 23:59:59',
            '2021-12-01 00:00:00', '2021-12-01 00:00:01',
            np.NaN,
            '2021-12-02 00:00:01', '2021-12-02 23:59:59',
        ])
        index_values = list(reversed(range(0, len(datetimes))))
        date_series = pd.Series(datetimes, index=index_values)
        expected_day = pd.Series(pd.to_datetime([
            '2021-01-01', '2021-01-01',
            np.NaN,
            '2021-01-02', '2021-01-02',
            '2021-02-01', '2021-02-01',
            np.NaN,
            '2021-02-02', '2021-02-02',
            '2021-03-01', '2021-03-01',
            np.NaN,
            '2021-03-02', '2021-03-02',
            '2021-04-01', '2021-04-01',
            np.NaN,
            '2021-04-02', '2021-04-02',
            '2021-05-01', '2021-05-01',
            np.NaN,
            '2021-05-02', '2021-05-02',
            '2021-06-01', '2021-06-01',
            np.NaN,
            '2021-06-02', '2021-06-02',
            '2021-07-01', '2021-07-01',
            np.NaN,
            '2021-07-02', '2021-07-02',
            '2021-08-01', '2021-08-01',
            np.NaN,
            '2021-08-02', '2021-08-02',
            '2021-09-01', '2021-09-01',
            np.NaN,
            '2021-09-02', '2021-09-02',
            '2021-10-01', '2021-10-01',
            np.NaN,
            '2021-10-02', '2021-10-02',
            '2021-11-01', '2021-11-01',
            np.NaN,
            '2021-11-02', '2021-11-02',
            '2021-12-01', '2021-12-01',
            np.NaN,
            '2021-12-02', '2021-12-02',
        ]), index=index_values)
        expected_month = pd.Series(pd.to_datetime([
            '2021-01-01', '2021-01-01',
            np.NaN,
            '2021-01-01', '2021-01-01',
            '2021-02-01', '2021-02-01',
            np.NaN,
            '2021-02-01', '2021-02-01',
            '2021-03-01', '2021-03-01',
            np.NaN,
            '2021-03-01', '2021-03-01',
            '2021-04-01', '2021-04-01',
            np.NaN,
            '2021-04-01', '2021-04-01',
            '2021-05-01', '2021-05-01',
            np.NaN,
            '2021-05-01', '2021-05-01',
            '2021-06-01', '2021-06-01',
            np.NaN,
            '2021-06-01', '2021-06-01',
            '2021-07-01', '2021-07-01',
            np.NaN,
            '2021-07-01', '2021-07-01',
            '2021-08-01', '2021-08-01',
            np.NaN,
            '2021-08-01', '2021-08-01',
            '2021-09-01', '2021-09-01',
            np.NaN,
            '2021-09-01', '2021-09-01',
            '2021-10-01', '2021-10-01',
            np.NaN,
            '2021-10-01', '2021-10-01',
            '2021-11-01', '2021-11-01',
            np.NaN,
            '2021-11-01', '2021-11-01',
            '2021-12-01', '2021-12-01',
            np.NaN,
            '2021-12-01', '2021-12-01',
        ]), index=index_values)
        expected_quarter = pd.Series(pd.to_datetime([
            '2021-01-01', '2021-01-01',
            np.NaN,
            '2021-01-01', '2021-01-01',
            '2021-01-01', '2021-01-01',
            np.NaN,
            '2021-01-01', '2021-01-01',
            '2021-01-01', '2021-01-01',
            np.NaN,
            '2021-01-01', '2021-01-01',
            '2021-04-01', '2021-04-01',
            np.NaN,
            '2021-04-01', '2021-04-01',
            '2021-04-01', '2021-04-01',
            np.NaN,
            '2021-04-01', '2021-04-01',
            '2021-04-01', '2021-04-01',
            np.NaN,
            '2021-04-01', '2021-04-01',
            '2021-07-01', '2021-07-01',
            np.NaN,
            '2021-07-01', '2021-07-01',
            '2021-07-01', '2021-07-01',
            np.NaN,
            '2021-07-01', '2021-07-01',
            '2021-07-01', '2021-07-01',
            np.NaN,
            '2021-07-01', '2021-07-01',
            '2021-10-01', '2021-10-01',
            np.NaN,
            '2021-10-01', '2021-10-01',
            '2021-10-01', '2021-10-01',
            np.NaN,
            '2021-10-01', '2021-10-01',
            '2021-10-01', '2021-10-01',
            np.NaN,
            '2021-10-01', '2021-10-01',
        ]), index=index_values)
        expected_quarter_fiscal_2 = pd.Series(pd.to_datetime([
            '2020-11-01', '2020-11-01',
            np.NaN,
            '2020-11-01', '2020-11-01',
            '2021-02-01', '2021-02-01',
            np.NaN,
            '2021-02-01', '2021-02-01',
            '2021-02-01', '2021-02-01',
            np.NaN,
            '2021-02-01', '2021-02-01',
            '2021-02-01', '2021-02-01',
            np.NaN,
            '2021-02-01', '2021-02-01',
            '2021-05-01', '2021-05-01',
            np.NaN,
            '2021-05-01', '2021-05-01',
            '2021-05-01', '2021-05-01',
            np.NaN,
            '2021-05-01', '2021-05-01',
            '2021-05-01', '2021-05-01',
            np.NaN,
            '2021-05-01', '2021-05-01',
            '2021-08-01', '2021-08-01',
            np.NaN,
            '2021-08-01', '2021-08-01',
            '2021-08-01', '2021-08-01',
            np.NaN,
            '2021-08-01', '2021-08-01',
            '2021-08-01', '2021-08-01',
            np.NaN,
            '2021-08-01', '2021-08-01',
            '2021-11-01', '2021-11-01',
            np.NaN,
            '2021-11-01', '2021-11-01',
            '2021-11-01', '2021-11-01',
            np.NaN,
            '2021-11-01', '2021-11-01',
        ]), index=index_values)

        self.assertEqual(
            expected_day.dt.date.tolist(),
            list(date.floor(date_series.tolist(), granularity=date.Granularity.DAY))
        )

        # without series.name
        self.assertTrue(validation.dataframes_match([
            pd.DataFrame(date_series.dt.date),
            pd.DataFrame(expected_day.dt.date),
            pd.DataFrame(date.floor(date_series, granularity=date.Granularity.DAY))
        ], ignore_indexes=False))

        self.assertTrue(validation.dataframes_match([
            pd.DataFrame(expected_month.dt.date),
            pd.DataFrame(date.floor(date_series, granularity=date.Granularity.MONTH))
        ], ignore_indexes=False))

        self.assertTrue(validation.dataframes_match([
            pd.DataFrame(expected_quarter.dt.date),
            pd.DataFrame(date.floor(date_series, granularity=date.Granularity.QUARTER))
        ], ignore_indexes=False))

        # with series.name
        date_series.name = 'date_day'
        actual_values = date.floor(date_series, granularity=date.Granularity.DAY)
        self.assertEqual(actual_values.name, 'date_day')
        self.assertTrue(all(actual_values.index.values == index_values))
        self.assertTrue(validation.dataframes_match([
            pd.DataFrame(expected_day.dt.date),
            pd.DataFrame(actual_values)
        ], ignore_indexes=False))

        date_series.name = 'date_month'
        actual_values = date.floor(date_series, granularity=date.Granularity.MONTH)
        self.assertEqual(actual_values.name, 'date_month')
        self.assertTrue(all(actual_values.index.values == index_values))
        self.assertTrue(validation.dataframes_match([
            pd.DataFrame(expected_month.dt.date),
            pd.DataFrame(actual_values)
        ], ignore_indexes=False))

        date_series.name = 'date_quarter'
        actual_values = date.floor(date_series, granularity=date.Granularity.QUARTER)
        self.assertEqual(actual_values.name, 'date_quarter')
        self.assertTrue(all(actual_values.index.values == index_values))
        self.assertTrue(validation.dataframes_match([
            pd.DataFrame(expected_quarter.dt.date),
            pd.DataFrame(actual_values)
        ], ignore_indexes=False))

        date_series.name = 'date_quarter_fiscal'
        actual_values = date.floor(
            date_series,
            granularity=date.Granularity.QUARTER,
            fiscal_start=2
        )
        self.assertEqual(actual_values.name, 'date_quarter_fiscal')
        self.assertTrue(all(actual_values.index.values == index_values))
        self.assertTrue(validation.dataframes_match([
            pd.DataFrame(expected_quarter_fiscal_2.dt.date),
            pd.DataFrame(actual_values)
        ], ignore_indexes=False))



import unittest
from datetime import timedelta

import numpy as np
import pandas as pd

from helpsk import string as hs
from helpsk import validation as hv
from helpsk import exceptions as he


# noinspection PyMethodMayBeStatic
class TestValidation(unittest.TestCase):

    def test_any_none_nan(self):
        self.assertTrue(hv.any_none_nan(None))
        self.assertTrue(hv.any_none_nan(np.NaN))
        self.assertTrue(hv.any_none_nan(pd.NaT))
        self.assertTrue(hv.any_none_nan(pd.NA))

        self.assertFalse(hv.any_none_nan(0))
        self.assertFalse(hv.any_none_nan(1))
        self.assertFalse(hv.any_none_nan(''))
        self.assertFalse(hv.any_none_nan('a'))

        # test list
        self.assertTrue(hv.any_none_nan([1, np.nan, None]))
        self.assertTrue(hv.any_none_nan([1, np.nan]))
        self.assertTrue(hv.any_none_nan([1, pd.NA]))
        self.assertTrue(hv.any_none_nan([1, pd.NaT]))
        self.assertTrue(hv.any_none_nan([1, None]))
        self.assertTrue(hv.any_none_nan([np.nan]))
        self.assertTrue(hv.any_none_nan([pd.NA]))
        self.assertTrue(hv.any_none_nan([pd.NaT]))
        self.assertTrue(hv.any_none_nan([None]))
        self.assertTrue(hv.any_none_nan([]))
        self.assertFalse(hv.any_none_nan([1]))
        self.assertFalse(hv.any_none_nan(['']))

        # test numpy array
        self.assertTrue(hv.any_none_nan(np.array([1, np.nan, None])))
        self.assertTrue(hv.any_none_nan(np.array([1, np.nan])))
        self.assertTrue(hv.any_none_nan(np.array([1, pd.NA])))
        self.assertTrue(hv.any_none_nan(np.array([1, pd.NaT])))
        self.assertTrue(hv.any_none_nan(np.array([1, None])))
        self.assertTrue(hv.any_none_nan(np.array([np.nan])))
        self.assertTrue(hv.any_none_nan(np.array([pd.NA])))
        self.assertTrue(hv.any_none_nan(np.array([pd.NaT])))
        self.assertTrue(hv.any_none_nan(np.array([None])))
        self.assertTrue(hv.any_none_nan(np.array([])))
        self.assertFalse(hv.any_none_nan(np.array([1])))
        self.assertFalse(hv.any_none_nan(np.array([''])))

        # test pandas series
        self.assertTrue(hv.any_none_nan(pd.Series([1, np.nan, None])))
        self.assertTrue(hv.any_none_nan(pd.Series([1, np.nan])))
        self.assertTrue(hv.any_none_nan(pd.Series([1, pd.NA])))
        self.assertTrue(hv.any_none_nan(pd.Series([1, pd.NaT])))
        self.assertTrue(hv.any_none_nan(pd.Series([1, None])))
        self.assertTrue(hv.any_none_nan(pd.Series([np.nan])))
        self.assertTrue(hv.any_none_nan(pd.Series([pd.NA])))
        self.assertTrue(hv.any_none_nan(pd.Series([pd.NaT])))
        self.assertTrue(hv.any_none_nan(pd.Series([None])))
        self.assertTrue(hv.any_none_nan(pd.Series([], dtype=float)))
        self.assertFalse(hv.any_none_nan(pd.Series([1])))
        self.assertFalse(hv.any_none_nan(pd.Series([''])))

        # test pandas data.frame
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[1, np.nan, None], [1, 2, 3]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[1, np.nan], [1, 2]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[1, pd.NA], [1, 2]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[1, pd.NaT], [1, 2]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[1, None], [1, 2]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[np.nan], [1]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[pd.NA], [1]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[pd.NaT], [1]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[None], [1]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([], dtype=float)))
        self.assertFalse(hv.any_none_nan(pd.DataFrame([1])))
        self.assertFalse(hv.any_none_nan(pd.DataFrame([[1], [1]])))
        self.assertFalse(hv.any_none_nan(pd.DataFrame([[''], [1]])))

    def test_any_missing(self):
        self.assertTrue(hv.any_missing(None))
        self.assertTrue(hv.any_missing(np.NaN))
        self.assertTrue(hv.any_missing(pd.NA))
        self.assertTrue(hv.any_missing(pd.NaT))
        self.assertTrue(hv.any_missing(''))

        self.assertFalse(hv.any_missing(0))

        # test list
        self.assertTrue(hv.any_missing([1, np.nan, None]))
        self.assertTrue(hv.any_missing([1, np.nan]))
        self.assertTrue(hv.any_missing([1, pd.NA]))
        self.assertTrue(hv.any_missing([1, pd.NaT]))
        self.assertTrue(hv.any_missing([1, None]))
        self.assertTrue(hv.any_missing([np.nan]))
        self.assertTrue(hv.any_missing([pd.NA]))
        self.assertTrue(hv.any_missing([pd.NaT]))
        self.assertTrue(hv.any_missing([None]))
        self.assertTrue(hv.any_missing([]))
        self.assertTrue(hv.any_missing(['']))
        self.assertTrue(hv.any_missing(['abc', '']))
        self.assertTrue(hv.any_missing([1, '']))
        self.assertFalse(hv.any_missing([1]))
        self.assertFalse(hv.any_missing(['a']))

        # test pandas series
        self.assertTrue(hv.any_missing(pd.Series([1, np.nan, None])))
        self.assertTrue(hv.any_missing(pd.Series([1, np.nan])))
        self.assertTrue(hv.any_missing(pd.Series([1, pd.NA])))
        self.assertTrue(hv.any_missing(pd.Series([1, pd.NaT])))
        self.assertTrue(hv.any_missing(pd.Series([1, None])))
        self.assertTrue(hv.any_missing(pd.Series([np.nan])))
        self.assertTrue(hv.any_missing(pd.Series([pd.NA])))
        self.assertTrue(hv.any_missing(pd.Series([pd.NaT])))
        self.assertTrue(hv.any_missing(pd.Series([None])))
        self.assertTrue(hv.any_missing(pd.Series([], dtype=float)))
        self.assertTrue(hv.any_missing(pd.Series([''])))
        self.assertTrue(hv.any_missing(pd.Series(['abc', ''])))
        self.assertTrue(hv.any_missing(pd.Series([1, ''])))
        self.assertFalse(hv.any_missing(pd.Series([1])))
        self.assertFalse(hv.any_missing(pd.Series(['a'])))

        # test pandas data.frame
        self.assertTrue(hv.any_missing(pd.DataFrame([[1, np.nan, None], [1, 2, 3]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[1, np.nan], [1, 2]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[1, pd.NA], [1, 2]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[1, pd.NaT], [1, 2]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[1, None], [1, 2]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[np.nan], [1]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[pd.NA], [1]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[pd.NaT], [1]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[None], [1]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([], dtype=float)))
        self.assertTrue(hv.any_missing(pd.DataFrame([['abc', ''], ['abc', 'abc']])))
        self.assertTrue(hv.any_missing(pd.DataFrame([''])))
        self.assertFalse(hv.any_missing(pd.DataFrame([1])))
        self.assertFalse(hv.any_missing(pd.DataFrame([[1], [1]])))

    def test_any_duplicated(self):
        # test list
        self.assertFalse(hv.any_duplicated(['']))
        self.assertFalse(hv.any_duplicated(['', 1]))
        self.assertFalse(hv.any_duplicated(['', 1, None]))

        self.assertTrue(hv.any_duplicated(['', 1, '']))
        self.assertTrue(hv.any_duplicated(['', 1, 1]))
        self.assertTrue(hv.any_duplicated(['', 1, None, None]))

        # test pd.Series
        self.assertFalse(hv.any_duplicated(pd.Series([''])))
        self.assertFalse(hv.any_duplicated(pd.Series(['', 1])))
        self.assertFalse(hv.any_duplicated(pd.Series(['', 1, None])))

        self.assertTrue(hv.any_duplicated(pd.Series(['', 1, ''])))
        self.assertTrue(hv.any_duplicated(pd.Series(['', 1, 1])))
        self.assertTrue(hv.any_duplicated(pd.Series(['', 1, None, None])))

    def test_iterables_are_equal(self):

        def subtest_iterables_are_equal(iterable_a, iterable_b, dtype):
            assert hv.iterables_are_equal(iterable_a, iterable_b)
            if dtype != str:
                assert hv.iterables_are_equal(np.array(iterable_a), np.array(iterable_b))
                assert hv.iterables_are_equal(np.array(iterable_a), iterable_b)
            assert hv.iterables_are_equal(
                pd.Series(iterable_a, dtype=dtype),
                pd.Series(iterable_b, dtype=dtype)
            )

        subtest_iterables_are_equal([], [], dtype=np.float64)
        subtest_iterables_are_equal([np.nan], [np.nan], np.float64)
        subtest_iterables_are_equal([1, 2, 3, 4], [1, 2, 3, 4], np.float64)
        subtest_iterables_are_equal([np.nan, 1, 2, 3, 4], [np.nan, 1, 2, 3, 4], np.float64)
        subtest_iterables_are_equal([np.nan, 1, 2, 3, 4, np.nan], [np.nan, 1, 2, 3, 4, np.nan], np.float64)  # noqa

        subtest_iterables_are_equal(iterable_a=['a', 'b', 'c', 'd'], iterable_b=['a', 'b', 'c', 'd'], dtype=str)  # noqa
        subtest_iterables_are_equal([np.nan, 'a', 'b', 'c', 'd'], [np.nan, 'a', 'b', 'c', 'd'], str)  # noqa
        subtest_iterables_are_equal([np.nan, 'a', 'b', 'c', 'd', np.nan], [np.nan, 'a', 'b', 'c', 'd', np.nan], str)  # noqa

        subtest_iterables_are_equal(['a', 'b', 'c', 4], ['a', 'b', 'c', 4], str)
        subtest_iterables_are_equal([np.nan, 'a', 'b', 'c', 4], [np.nan, 'a', 'b', 'c', 4], str)
        subtest_iterables_are_equal([np.nan, 'a', 'b', 'c', 4, np.nan], [np.nan, 'a', 'b', 'c', 4, np.nan], str)  # noqa

        def subtest_iterables_are_not_equal(iterable_a, iterable_b, dtype):
            assert not hv.iterables_are_equal(iterable_a, iterable_b)
            assert not hv.iterables_are_equal(np.array(iterable_a), np.array(iterable_b))
            assert not hv.iterables_are_equal(pd.Series(iterable_a, dtype=dtype), pd.Series(iterable_b, dtype=dtype))  # noqa

        subtest_iterables_are_not_equal(iterable_a=[], iterable_b=[1], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[1], iterable_b=[], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[], iterable_b=[np.nan], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[np.nan], iterable_b=[], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[], iterable_b=[''], dtype=str)
        subtest_iterables_are_not_equal(iterable_a=[''], iterable_b=[], dtype=str)

        subtest_iterables_are_not_equal(iterable_a=[1, 2, 3], iterable_b=[1, 2, 3, 4], dtype=np.float64)  # noqa
        subtest_iterables_are_not_equal(iterable_a=[0, 1, 2, 3], iterable_b=[1, 2, 3], dtype=np.float64)  # noqa
        subtest_iterables_are_not_equal(iterable_a=[1.000001, 2, 3], iterable_b=[1, 2, 3], dtype=np.float64)  # noqa
        subtest_iterables_are_not_equal(iterable_a=[1, 2, 3], iterable_b=[1.000001, 2, 3], dtype=np.float64)  # noqa
        subtest_iterables_are_not_equal(iterable_a=[1, 2, 3], iterable_b=[1.000001, 2, 3], dtype=np.float64)  # noqa
        subtest_iterables_are_not_equal(iterable_a=[1, 2, 3], iterable_b=[np.nan, 2, 3], dtype=np.float64)  # noqa
        subtest_iterables_are_not_equal(iterable_a=[np.nan, 2, 3], iterable_b=[1, 2, 3], dtype=np.float64)  # noqa
        subtest_iterables_are_not_equal(iterable_a=[1, 2, 3], iterable_b=[None, 2, 3], dtype=np.float64)  # noqa

    def test_iterables_are_equal__ordered(self):
        # found a bug where two pd.Categorical series will not return True even if they have the
        # same values, but with different ordered; but we only care about the values
        values = [np.nan, 'a', 'b', 'c', 'd']
        series = pd.Series(values)
        unordered_categorical = pd.Categorical(values=values, ordered=False)
        ordered_categorical = pd.Categorical(values=values, categories=['d', 'c', 'a', 'b'], ordered=True)  # noqa

        self.assertTrue(hv.iterables_are_equal(values, series))
        self.assertTrue(hv.iterables_are_equal(values, unordered_categorical))
        self.assertTrue(hv.iterables_are_equal(values, ordered_categorical))
        self.assertTrue(hv.iterables_are_equal(series, unordered_categorical))
        self.assertTrue(hv.iterables_are_equal(series, ordered_categorical))
        self.assertTrue(hv.iterables_are_equal(unordered_categorical, ordered_categorical))

        different_values = values.copy()
        different_values[-1] = 'z'

        self.assertFalse(hv.iterables_are_equal(different_values, series))
        self.assertFalse(hv.iterables_are_equal(different_values, unordered_categorical))
        self.assertFalse(hv.iterables_are_equal(different_values, ordered_categorical))
        self.assertFalse(hv.iterables_are_equal(pd.Categorical(different_values, ordered=False),
                                                unordered_categorical))
        self.assertFalse(hv.iterables_are_equal(pd.Categorical(different_values, ordered=True),
                                                ordered_categorical))

    def test_dataframes_match(self):

        dataframe_1 = pd.DataFrame({
            'col_floats': [1.123456789, 2.123456789, 3.123456789, np.nan],
            'col_strings': [np.nan, 'a', 'b', 'c'],
            'col_enums': [np.nan, hs.RoundTo.NONE, hs.RoundTo.AUTO, hs.RoundTo.THOUSANDS],
            'col_dates': pd.date_range('2021-01-01', periods=4),
            'col_missing': [np.nan, np.nan, np.nan, np.nan]}
        )
        dataframe_1.loc[0, 'col_dates'] = np.nan

        with self.assertRaises(he.HelpskParamTypeError):
            hv.dataframes_match(dataframes=None)  # noqa

        with self.assertRaises(he.HelpskParamTypeError):
            hv.dataframes_match(dataframes=dataframe_1)  # noqa

        with self.assertRaises(he.HelpskParamValueError):
            hv.dataframes_match(dataframes=[])

        with self.assertRaises(he.HelpskParamValueError):
            hv.dataframes_match(dataframes=[dataframe_1])

        # test that there are no side effects; e.g. we set index/column values if we ignore them
        dataframe_1_original = dataframe_1.copy()
        dataframe_2 = dataframe_1.copy()
        dataframe_2.index = ['a', 'b', 'c', 'd']
        dataframe_2.columns = ['a', 'b', 'c', 'd', 'e']
        dataframe_2_original = dataframe_2.copy()
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2],
                                            float_tolerance=1,
                                            ignore_indexes=True,
                                            ignore_column_names=True))
        self.assertTrue(dataframe_1.equals(dataframe_1_original))
        self.assertTrue((dataframe_1.index == dataframe_1_original.index).all())
        self.assertTrue((dataframe_1.columns == dataframe_1_original.columns).all())
        self.assertTrue(dataframe_2.equals(dataframe_2_original))
        self.assertTrue((dataframe_2.index == dataframe_2_original.index).all())
        self.assertTrue((dataframe_2.columns == dataframe_2_original.columns).all())

        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1]))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1, dataframe_1]))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_1.copy()]))  # noqa

        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.round(6)]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.round(5)]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1, dataframe_1.round(5)]))  # noqa
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.round(5)],
                                            float_tolerance=5))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.round(6),
                                                        dataframe_1.round(5)],
                                            float_tolerance=5))

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[0, 0]))
        dataframe_2.iat[0, 0] = dataframe_2.iat[0, 0] - 0.000001
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))  # noqa

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[1, 1]))
        dataframe_2.iat[1, 1] = 'c'
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))  # noqa

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[2, 2]))
        dataframe_2.iat[2, 2] = hs.RoundTo.MILLIONS
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))  # noqa

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[2, 3]))
        dataframe_2.iat[2, 3] = dataframe_2.iat[2, 3] + timedelta(seconds=1)
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))  # noqa

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[2, 3]))
        dataframe_2.iat[2, 3] = dataframe_2.iat[2, 3] + timedelta(seconds=1) - timedelta(seconds=1)
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))  # noqa

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[0, 0]))
        dataframe_2.iat[0, 0] = np.NaN
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))  # noqa

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[1, 1]))
        dataframe_2.iat[1, 1] = np.NaN
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))  # noqa

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[1, 2]))
        dataframe_2.iat[1, 2] = np.NaN
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))  # noqa

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[1, 3]))
        dataframe_2.iat[1, 3] = np.NaN
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))  # noqa

        dataframe_2 = dataframe_1.copy()
        dataframe_2.index = ['a', 'b', 'c', 'd']
        dataframe_2.columns = ['a', 'b', 'c', 'd', 'e']
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2], ignore_indexes=False))  # noqa
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2],
                                             ignore_column_names=False))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1],
                                             ignore_column_names=False))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2],
                                             ignore_indexes=False,
                                             ignore_column_names=False))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1],
                                             ignore_indexes=False,
                                             ignore_column_names=False))

    def test_is_close(self):
        self.assertFalse(hv.is_close(np.nan, 1))
        self.assertFalse(hv.is_close(1, np.nan))
        self.assertFalse(hv.is_close(np.nan, np.nan))
        with self.assertRaises(Exception):
            hv.is_close(None, 1)  # noqa
        with self.assertRaises(Exception):
            hv.is_close(1, None)  # noqa
        with self.assertRaises(Exception):
            hv.is_close(None, None)  # noqa

        def sub_test(value, tolerance, failure_tolerance):
            self.assertTrue(hv.is_close(value, value))
            self.assertTrue(hv.is_close(value, value + tolerance))
            self.assertTrue(hv.is_close(value + tolerance, value))
            self.assertTrue(hv.is_close(-value, -value - tolerance))
            self.assertTrue(hv.is_close(-value - tolerance, -value))

        sub_test(value=0, tolerance=0.000001, failure_tolerance=0.0000011)
        sub_test(value=1, tolerance=0.000001, failure_tolerance=0.0000011)
        sub_test(value=100, tolerance=0.000001, failure_tolerance=0.0000011)
        sub_test(value=-100, tolerance=0.000001, failure_tolerance=0.0000011)

    def test_raises_exception(self):

        def my_function_exception():
            raise ValueError()

        def my_function_runs():
            return True

        self.assertTrue(hv.raises_exception(my_function_exception))
        # should return True since my_function_exception raises ValueError
        self.assertTrue(hv.raises_exception(my_function_exception, ValueError))
        # should return False since my_function_exception raises ValueError, not TypeError
        self.assertFalse(hv.raises_exception(my_function_exception, TypeError))
        self.assertFalse(hv.raises_exception(my_function_runs))



"""
Contains the base classes that wraps the connection/querying logic of
various databases.

See documentation in database.py
"""

from abc import ABC, abstractmethod
from typing import TypeVar
import configparser
import time
import pandas as pd
from helpsk.utility import suppress_stdout, suppress_warnings


ConnectionObject = TypeVar('ConnectionObject')


class Database(ABC):
    """Wraps the connection/querying logic of various databases."""

    def __init__(self, **kwargs: dict):
        """
        **kwargs should contain the names of the underlying connection object and corresponding
        values.

        For example:
            - For Redshift, typical keyword arguments might be:
                - `dbname`, `password`, `database`, `port`, `host`
            - For Snowflake, typical keyword arguments might be:
                - 'user', 'account', 'authenticator', 'warehouse', 'database', 'autocommit'
        """
        self._kwargs = kwargs
        self.connection_object: ConnectionObject = None

    @classmethod
    def from_config(cls, config_path: str, config_key: str) -> 'Database':  # noqa: ANN102
        """
        Passes key/value pairs found in configuration file into underlying connection. Keys must
        match corresponding connection string or connection method arguments.

        For example:
            - For Redshift, typical keys might be:
                - `dbname`, `password`, `database`, `port`, `host`

            ```
            [redshift]
            user=my_username
            password=my-password-123
            dbname=the_database
            host=host.address.redshift.amazonaws.com
            port=1234
            ```

            - For Snowflake, typical keys might be:
                - 'user', 'account', 'authenticator', 'warehouse', 'database', 'autocommit'

            ```
            [snowflake]
            user=my.email@address.com
            account=account.id
            authenticator=externalbrowser
            warehouse=WAREHOUSE_NAME
            database=DATABASE_NAME
            ```

        Args:
            config_path:
                path to the configuration file containing key/value pairs.
            config_key:
                name of the configuration key in configuration file
                e.g. `[redshift]`
        """
        config = configparser.ConfigParser()
        config.read(config_path)
        config_dict = dict(config[config_key].items())
        return cls(**config_dict)

    def is_connected(self) -> bool:
        """Returns True if the database is connected, otherwise False."""
        return self.connection_object is not None

    @abstractmethod
    def _open_connection_object(self) -> ConnectionObject:
        """
        Child classes will implement the logic to connect to the database and return the
        connection object.

        The returning value will be stored in self.connection_object

        Returns the connection object (usually returned from the underlying `connect()` method.
        For example, the returning value from `snowflake.connection.connect()`
        """

    @abstractmethod
    def _close_connection_object(self) -> None:
        """
        Child classes will implement the logic to close the connection to the database,
        typically by calling `close()` on the connection object returned from
        _open_connection_object.

        For example:
            self.connection_object.close()
        """

    def connect(self) -> None:
        """
        Call this method to open the connection to the database.

        Alternatively you can use the context manager:

        ```
        with Database(...) as database:
            database.query("SELECT * FROM table LIMIT 100")
        ```
        """
        if not self.is_connected():
            self.connection_object = self._open_connection_object()
            assert self.is_connected()

    def close(self) -> None:
        """
        Call this method to close the connection to the database.

        Alternatively you can use the context manager:

        ```
        with Database(...) as database:
            database.query("SELECT * FROM table LIMIT 100")
        ```
        """
        if self.is_connected():
            self._close_connection_object()
            self.connection_object = None
            assert not self.is_connected()

    def __enter__(self):
        """Enables the context manager."""
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):  # noqa: ANN001
        """Enables the context manager."""
        self.close()

    @abstractmethod
    def _query(self, sql: str) -> pd.DataFrame:
        """
        Method for child classes to override that contains logic to query.

        Args:
            sql:
                SQL to execute e.g. "SELECT * FROM table LIMIT 100"

        Returns:
            a pandas Dataframe with the results from the query
        """

    def query(self, sql: str, *,
              show_messages: bool = False,
              show_warnings: bool = False,
              show_elapsed_time: bool = True) -> pd.DataFrame:
        """
        Queries the database and returns the results as a pandas Dataframe.

        Args:
            sql:
                SQL to execute e.g. "SELECT * FROM table LIMIT 100"
            show_messages:
                if True, shows messages generated from the underlying query
            show_warnings:
                if True, shows warnings generated from the underlying query
            show_elapsed_time:
                if True, shows the elapsed execution time for the query

        Returns:
            a pandas Dataframe with the results from the query.
        """
        assert self.is_connected()
        start_time = time.time()
        if show_messages and show_warnings:
            with suppress_stdout(), suppress_warnings():
                results = self._query(sql=sql)
        elif show_messages:
            with suppress_stdout():
                results = self._query(sql=sql)
        elif show_warnings:
            with suppress_warnings():
                results = self._query(sql=sql)
        else:
            results = self._query(sql=sql)

        end_time = time.time()

        if show_elapsed_time:
            elapsed_time = end_time - start_time
            time_units = 'minutes' if elapsed_time > 60 else 'seconds'
            elapsed_time = elapsed_time / 60 if elapsed_time > 60 else elapsed_time
            elapsed_time = round(elapsed_time, 1)
            print(f'Elapsed Time: {elapsed_time} {time_units}')

        return results

    @abstractmethod
    def execute_statement(self, statement: str) -> None:
        """Executes a statement without any data returned."""

    @abstractmethod
    def insert_records(
        self,
        dataframe: pd.DataFrame,
        table: str,
        create_table: bool = False,
        overwrite: bool = False,
        schema: str | None = None,
        database: str | None = None) -> None:
        """
        Inserts rows into a table from a pandas DataFrame.

        Args:
            dataframe:
                the pandas dataframe to insert
            table:
                the name of the table
            create_table:
                if True, creates the table before inserting
            overwrite:
                if True, drops all records before inserting
            schema:
                the name of the schema
            database:
                the name of the database
        """



"""A collection of functions that assist in validation/comparison of data and conditions."""

from __future__ import annotations
from collections.abc import Sized, Iterable, Callable

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_categorical_dtype

from helpsk.exceptions import HelpskParamTypeError, HelpskParamValueError
from helpsk.utility import suppress_warnings


def is_none_nan(value: object) -> bool:
    """Returns True if the value is None or various NaN/NA types."""
    if value is None \
            or value is pd.NA \
            or value is pd.NaT \
            or (isinstance(value, (np.float64, float, int)) and np.isnan(value)):
        return True

    return False


def any_none_nan(  # noqa: PLR0911
        values: list | np.ndarray | pd.Series | pd.DataFrame | object) -> bool:
    """
    Can be used with a single value or a collection of values. Returns `True` if any item in
    `values` are `None`, `np.Nan`, `pd.NA`, `pd.NaT` or if the length of `values` is `0`.

    Args:
        values:
            A collection of values to check.

    Returns:
        bool - True if any item in `values` are None/np.NaN
    """
    if is_none_nan(values):
        return True

    if isinstance(values, Sized) and not isinstance(values, str) and len(values) == 0:
        return True

    if isinstance(values, pd.Series):
        return values.isna().any() or values.isna().any()

    if isinstance(values, pd.DataFrame):
        return values.isna().any().any() or values.isna().any().any()

    if isinstance(values, Iterable) and not isinstance(values, str):
        if len(values) == 0:
            return True

        return any(is_none_nan(x) for x in values)

    try:
        if not isinstance(values, str) and None in values:
            return True
    except Exception:
        pass

    try:
        if np.isnan(values).any():
            return True
    except TypeError:
        return False

    return False


def any_missing(values: list | pd.Series | pd.DataFrame | object) -> bool:
    """
    Same as `any_none_nan` but checks for empty strings.

    Args:
        values:
            A collection of values to check.

    Returns:
        bool - True if any item in `values` are None/np.NaN/''
    """
    if any_none_nan(values):
        return True

    if isinstance(values, pd.Series):
        return values.isin(['']).any()

    if isinstance(values, pd.DataFrame):
        return values.isin(['']).any().any()

    if isinstance(values, str) and values.strip() == '':
        return True

    if isinstance(values, Iterable) and '' in values:
        return True

    return False


def any_duplicated(values: Iterable) -> bool:
    """
    Returns `True` if any items in `values` are duplicated.

    Args:
        values: list, np.ndarray, pd.Series
            A collection of values to check.

    Returns:
        bool
    """
    return len(values) != len(set(values))


def iterables_are_equal(iterable_a: Iterable, iterable_b: Iterable) -> bool:
    """
    Compares the equality of the values of two iterables.

    This function will generally give the same result as list equality (e.g.
    `[x, y, z] == [x, y, z]`). However, in some strange scenarios, `==` will return `False` where
    it doesn't seem like it should

    For example:

    ```
    temp = pd.DataFrame({'col_a': [np.nan, 1.0]})
    temp.col_a.tolist() == [np.nan, 1.0]  # returns False. Why??
    iterables_are_equal(temp.col_a.tolist(), [np.nan, 1])  # returns True
    [np.nan, 1.0] == [np.nan, 1.0]  # returns True


    Also, when comparing a series with an ordered Categorical when the values are the same,
    pd.Series.equals() will return False if the categories have different order. But we only care
    if the values are the same, so this function will return True.
    ```

    Args:
        iterable_a:
            an iterable to equate to iterable_b
        iterable_b:
            an iterable to equate to iterable_a

    Returns:
        True if iterable_a is equal to iterable_b
    """
    with suppress_warnings():
        # if either list-like structure is categorical, then we need to convert both to unordered
        # categorical
        if len(iterable_a) != len(iterable_b):
            return False
        if len(iterable_a) == 0 and len(iterable_b) == 0:
            return True

        if is_categorical_dtype(iterable_a) or is_categorical_dtype(iterable_b):
            iterable_a = pd.Categorical(iterable_a, ordered=False)
            iterable_b = pd.Categorical(iterable_b, ordered=False)
        else:
            iterable_a = pd.Series(iterable_a)
            iterable_b = pd.Series(iterable_b)

        return iterable_a.equals(iterable_b)


def dataframes_match(dataframes: list[pd.DataFrame],
                     float_tolerance: int = 6,
                     ignore_indexes: bool = True,
                     ignore_column_names: bool = True) -> bool:
    """
    Because floating point numbers are difficult to accurate represent, when comparing multiple
    DataFrames, this function first rounds any numeric columns to the number of decimal points
    indicated `float_tolerance`.

    Args:
        dataframes:
            Two or more dataframes to compare against each other and test for equality

        float_tolerance:
            numeric columns will be rounded to the number of digits to the right of the decimal
            specified by this parameter.

        ignore_indexes:
            if True, the indexes of each DataFrame will be ignored for considering equality

        ignore_column_names:
            if True, the column names of each DataFrame will be ignored for considering equality

    Returns:
        Returns True if the dataframes match based on the conditions explained above, otherwise
        returns False
    """
    if not isinstance(dataframes, list):
        raise HelpskParamTypeError("Expected list of pd.DataFrame's.")

    if not len(dataframes) >= 2:
        raise HelpskParamValueError("Expected 2 or more pd.DataFrame's in list.")

    first_dataframe = dataframes[0].round(float_tolerance)

    def first_dataframe_equals_other(other_dataframe: pd.DataFrame) -> bool:
        if first_dataframe.shape != other_dataframe.shape:
            return False

        if ignore_indexes or ignore_column_names:
            # if either of these are True, then we are going to change the index and/or columns,
            # but python is pass-by-reference so we don't want to change the original DataFrame
            # object.
            other_dataframe = other_dataframe.copy()

        if ignore_indexes:
            other_dataframe.index = first_dataframe.index

        if ignore_column_names:
            other_dataframe.columns = first_dataframe.columns

        return first_dataframe.equals(other_dataframe.round(float_tolerance))

    # compare the first dataframe to the rest of the dataframes, after rounding each to the
    # tolerance, and performing other modifications
    # check if all results are True
    return all(first_dataframe_equals_other(x) for x in dataframes[1:])


def is_close(value_a: float, value_b: float, tolerance: float = 0.000001) -> bool:
    """
    Tests whether or not value_a and value_b are "close" (i.e. within the `tolerance` after
    subtracting).

    Args:
        value_a:
            numeric value to test
        value_b:
            numeric value to test
        tolerance:
            the maximum difference (absolute value) allowed between value_a and value_b
    Returns:
          True if values are within specified tolerance
    """
    return abs(value_a - value_b) <= tolerance


def raises_exception(function: Callable, exception_type: type | None = None) -> bool:
    """
    Returns True if `function` raises an Exception; returns False if `function` runs without
    raising an Exception.

    Args:
        function:
            the function which does or does not raise an exception.
        exception_type:
            if `exception_type` is provided, `raises_exception` returns true only if the `function`
            argument raises an Exception **and** the exception has type of `exception_type`.
    """
    try:
        function()
        return False
    except Exception as exception:
        if exception_type:
            return isinstance(exception, exception_type)

        return True
